#!/usr/bin/env python3
# coding: utf-8

import io
import sys
import re

import tqdm
import marcx
import collections

formatmap = {
    "p||||||z|||||||": "Zeitschrift",
    "z||||||z|||||||": "Zeitung",
    "r||||||z|||||||": "Reihe",
    "r||||||||||||||": "Reihe",
    "a|a|||||||||||": "Monografie",
    "a|||||||||||||": "Monografie",
    "a|a|||||||": "Monografie",
    "|||||ca||||||": "Videokassette",
    "a|||||||g|||||": "Datenträger",
    "||||||||g|": "Datenträger"
}

def get_field(tag):
    regexp = re.search('<.*?\snr="%s".*>(.*)$' % tag, field)
    if regexp:
        _field = regexp.group(1)
        return _field
    else:
        return ""

def get_subfield(tag, subfield):
    regexp = re.search('<feld.*nr="%s".*><uf code="%s">(.*?)<\/uf>' % (tag, subfield), field)
    if regexp:
        _field = regexp.group(1)
        return _field
    else:
        return ""

inputfilename = "130_input.xml"
outputfilename = "130_output.mrc"

if len(sys.argv) == 3:
    inputfilename, outputfilename = sys.argv[1:]

hierarchymap = collections.defaultdict(list)

with io.open(inputfilename, "r") as inputfile:

    records = inputfile.read()
    records = records.split("</datensatz>")

    for record in records:

        f010 = ""
        f089 = ""

        record = record.replace("\n", "")
        pattern = """<feld nr="010" ind=" ">(.*?)</feld>.*?nr="089" ind=" ">(.*?)</feld>"""
        regexp = re.search(pattern, record)
        if regexp:
            f010, f089 = regexp.groups()

        if f010 != "" and f089 != "":
            f010 = f010.lstrip("0")
            hierarchymap[f010].append(f089)

    #print(hierarchymap)

if len(sys.argv) == 3:
    inputfilename, outputfilename = sys.argv[1:]

inputfile = io.open(inputfilename, "r")
outputfile = io.open(outputfilename, "wb")

records = inputfile.read()
records = records.split("</datensatz>")

for record in records:

    f001 = ""
    f007 = ""
    f020a = ""
    f041a = ""
    f100a = ""
    f100e = ""
    f110a = ""
    f245a = ""
    f245b = ""
    f245c = ""
    f250a = ""
    f260a = ""
    f260b = ""
    f260c = ""
    f300a = ""
    f300b = ""
    f300c = ""
    f500a = ""
    f650a = ""
    f700e = ""
    f856u = ""
    f866a = ""
    format1 = ""
    format2 = ""
    subjects = []
    persons = []
    corporates = []

    marcrecord = marcx.Record(force_utf8=True, to_unicode=True)
    marcrecord.strict = False
    fields = re.split("(</feld>)", record)

    for field in fields:

        field = field.replace("\n", "")
        field = field.replace("</feld>", "")
        field = field.replace("      ", "")

        # Identfikator
        if f001 == "":
            f001 = get_field("001")

        # ISBN
        if f020a == "":
            f020a = get_field("540")
            if f020a != "":
                regexp = re.search("\s?([\d-]+)\s?", f020a)
                if regexp:
                    f020a = regexp.group(1)
                else:
                    print("Die ISBN konnte nicht bereinigt werden: " + f020a)

        # Sprache
        if f041a == "":
            f041a = get_field("037")

        # 1. Urheber
        if f100a == "":
            f100a = get_field("100")
            if f100a != "":
                regexp = re.search(".\s¬(\[.*\])", f100a)
                if regexp:
                    f100e = regexp.group(1)
                    f100a = re.sub(".\s¬\[.*\]¬", "", f100a)

        # 1. Körperschaft
        if f110a == "":
            f110a = get_field("200")

        # Haupttitel
        if f245a == "":
            f245a = get_field("331")

        # Titelzusatz
        if f245b == "":
            f245b = get_field("335")

        # Verantwortlichenangabe
        if f245c == "":
            f245c = get_field("359")

        # Ausgabe
        if f250a == "":
            f250a = get_field("403")

        # Erscheinungsort
        if f260a == "":
            f260a = get_field("410")
            if f260a == "":
                f260a = get_subfield("419", "a")

        # Verlag
        if f260b == "":
            f260b = get_field("412")
            if f260b == "":
                f260b = get_subfield("419", "b")
            if f260b != "":
                f260b = " : " + f260b

        # Erscheinungsjahr
        if f260c == "":
            f260c = get_field("425")
            if f260c == "":
                f260c = get_subfield("419", "c")
            if f260c != "" and (f260a != "" or f260b != ""):
                f260c = ", " + f260c

        # Seitenzahl
        if f300a == "":
            f300a = get_field("433")

        # Illustrationen
        if f300b == "":
            f300b = get_field("434")
            if f300b != "":
                f300b = " : " + f300b

        # Format
        if f300c == "":
            f300c = get_field("435")
            if f300c != "":
                f300c = " ; " + f300c

        # Schlagwörter
        if f650a == "":
            f650a = get_field("710")
            if f650a != "":
                subjects.append(f650a)
                f650a = ""

        # weitere Personen
        regexp = re.search('nr="1\d\d"', field)
        if regexp:
            for i in range(101, 197):  # überprüfen, ob ein Personenfeld vorliegt, damit die Schleife für die Personenfelder nicht bei jedem Feld durchlaufen wird
                f700a = get_field(i)
                if f700a != "":
                    regexp = re.search("^\d+", f700a) # die Felder, die nur Personen-IDs enthalten, werden übersprungen
                    if not regexp:
                        regexp = re.search(".\s¬(\[.*\])", f700a)
                        if regexp:
                            f700e = regexp.group(1)
                            f700a = re.sub(".\s¬\[.*\]¬", "", f700a)
                        persons.append(f700a)
                        break

        # weitere Körperschaften
        regexp = re.search('nr="2\d\d"', field)
        if regexp:
            for i in range(201, 299):
                f710a = get_field(i)
                if f710a != "":
                    regexp = re.search("^\d+", f710a)
                    if not regexp:
                        corporates.append(f710a)
                        break

        # Link zum Datensatz
        if f856u == "":
            f856u = get_subfield("655", "u")

        #Bestandsnachweis (866a)
        f866a = f001.lstrip("0")
        f866a = hierarchymap.get(f866a, "")

        if format1 == "":
            format1 = get_field("052")

        if format2 == "":
            format2 = get_field("050")

    if format1 != "" or format2 != "":
        format = format1 or format2
    else:
        format = "a|a|||||||||||"

    if formatmap.get(format) == "Zeitschrift":
        leader = "     cas a2202221   4500"
        f007 = "tu"
        f008 = "                     p"
        f935b = "druck"
        f935c = "az"
    elif formatmap.get(format) == "Zeitung":
        leader = "     cas a2202221   4500"
        f007 = "tu"
        f008 = "                     n"
        f935b = "druck"
        f935c = "az"
    elif formatmap.get(format) == "Reihe":
        leader = "     cas a2202221   4500"
        f007 = "tu"
        f008 = "                     m"
        f935b = "druck"
        f935c = "az"
    elif formatmap.get(format) == "Monografie":
        leader = "     nam  22        4500"
        f007 = "tu"
        f008 = ""
        f935b = "druck"
        f935c = ""
    elif formatmap.get(format) == "Videokassette":
        leader = "     cgm  22        4500"
        f007 = "v"
        f008 = ""
        f935b = "vika"
        f935c = "vide"
    elif formatmap.get(format) == "Datenträger":
        leader = "     cgm  22        4500"
        f007 = "v"
        f008 = ""
        f935b = "soerd"
        f935c = ""
    else:
        print("Format %s ist nicht in der Mapping-Tabelle enthalten" % format)
        leader = "     nam  22        4500"
        f007 = "tu"
        f008 = ""
        f935b = "druck"
        f935c = ""

    if f245a == "" or "Arkady" in f245a: # einzelne Zeitschriftenhefte und die fehlerhaften Arkady-Records werden übersprungen
        continue

    assert(len(leader) == 24)
    marcrecord.leader = leader
    marcrecord.add("001", data="finc-130-" + f001)
    marcrecord.add("007", data=f007)
    marcrecord.add("008", data=f008)
    marcrecord.add("020", a=f020a)
    marcrecord.add("041", a=f041a)
    marcrecord.add("100", a=f100a, e=f100e)
    marcrecord.add("110", a=f110a)
    titleparts = ["a", f245a, "b", f245b, "c", f245c]
    marcrecord.add("245", subfields=titleparts)
    marcrecord.add("250", a=f250a)
    publisher = ["a", f260a, "b", f260b, "c", f260c]
    marcrecord.add("260", subfields=publisher)
    physicaldescription = ["a", f300a, "b", f300b, "c", f300c]
    marcrecord.add("300", subfields=physicaldescription)
    for subject in subjects:
        marcrecord.add("650", a=subject)
    for person in persons:
        marcrecord.add("700", a=person, e=f700e)
    for corporate in corporates:
        marcrecord.add("710", a=corporate)
    if f856u != "":
        marcrecord.add("856", q="text/html", u=f856u)
    f866a = "; ".join(f866a)
    marcrecord.add("866", a=f866a)
    marcrecord.add("935", b=f935b, c=f935c)
    marcrecord.add("980", a=f001, b="130", c="VDEH")

    outputfile.write(marcrecord.as_marc())

inputfile.close()
outputfile.close()



# coding: utf-8
# pylint: disable=C0301,E1101,C0103

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

#
# Aggregated Index workflows
# ==========================
#
# AMSL (api) + AIIntermediateSchema (sources) = AILicensing (isil) -> AIExport (for solr)
#

import binascii
import collections
import datetime
import itertools
import multiprocessing
import os
import re
import shutil
import string
import tempfile
import urllib

import luigi
import rdflib
import requests
import ujson as json

from bs4 import BeautifulSoup
from dateutil.relativedelta import relativedelta
from gluish.common import Executable
from gluish.format import TSV, Gzip
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.benchmark import timed
from siskin.database import sqlitedb
from siskin.sources.amsl import (AMSLFilterConfigFreeze, AMSLHoldingsFile,
                                 AMSLOpenAccessKBART, AMSLService, AMSLFreeContent)
from siskin.sources.arxiv import ArxivIntermediateSchema
from siskin.sources.ceeol import CeeolJournalsDumpIntermediateSchema
from siskin.sources.crossref import (CrossrefDOIList,
                                     CrossrefIntermediateSchema,
                                     CrossrefUniqISSNList)
from siskin.sources.degruyter import (DegruyterDOIList,
                                      DegruyterIntermediateSchema,
                                      DegruyterISSNList)
from siskin.sources.doaj import (DOAJDOIList, DOAJIntermediateSchema,
                                 DOAJISSNList)
from siskin.sources.elsevierjournals import (ElsevierJournalsIntermediateSchema,
                                             ElsevierJournalsISSNList)
from siskin.sources.genderopen import GenderopenIntermediateSchema
from siskin.sources.genios import (GeniosCombinedIntermediateSchema,
                                   GeniosISSNList)
from siskin.sources.hhbd import HHBDIntermediateSchema
from siskin.sources.ieee import IEEEDOIList, IEEEIntermediateSchema
from siskin.sources.ijoc import IJOCIntermediateSchema
from siskin.sources.jstor import (JstorDOIList, JstorIntermediateSchema,
                                  JstorISSNList)
from siskin.sources.kielfmf import KielFMFIntermediateSchema
from siskin.sources.lynda import LyndaIntermediateSchema
from siskin.sources.mag import MAGReferenceDB
from siskin.sources.pqdt import PQDTIntermediateSchema
from siskin.sources.springer import SpringerIntermediateSchema
from siskin.sources.ssoar import SSOARIntermediateSchema
from siskin.sources.thieme import ThiemeIntermediateSchema, ThiemeISSNList
from siskin.task import DefaultTask
from siskin.utils import URLCache, load_set_from_target


class AITask(DefaultTask):
    """ AI base task. """
    TAG = 'ai'

    def closest(self):
        return weekly(self.date)


class AIDOIList(AITask):
    """
    List of DOI for ai.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return [
            CrossrefDOIList(date=self.date),
            DegruyterDOIList(date=self.date),
            DOAJDOIList(date=self.date),
            ElsevierJournalsISSNList(date=self.date),
            IEEEDOIList(date=self.date),
            JstorDOIList(date=self.date),
        ]

    def run(self):
        """
        Concatenate files and run sort -u.
        """
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for target in self.input():
            shellout("cat {input} >> {output}",
                     input=target.path, output=stopover)
        output = shellout(
            "LC_ALL=C sort -S35% {input} > {output}", input=stopover)
        os.remove(stopover)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class AIDOIRedirectTable(AITask):
    """
    Generate a redirect table. Takes days. Make sure doi.org is in your hosts file so DNS is not stressed.
    """
    hurrly_workers = luigi.IntParameter(
        default=4, description='number of workers for hurrly')

    def requires(self):
        return AIDOIList()

    def run(self):
        output = shellout("hurrly -w {w} < {input} > {output}",
                          w=self.hurrly_workers, input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class AIDOIStats(AITask):
    """
    DOI overlaps.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'crossref': CrossrefDOIList(date=self.date),
            'degruyter': DegruyterDOIList(date=self.date),
            'doaj': DOAJDOIList(date=self.date),
            'jstor': JstorDOIList(date=self.date)
        }

    @timed
    def run(self):
        with self.output().open('w') as output:
            for k1, k2 in itertools.combinations(list(self.input().keys()), 2):
                s1 = load_set_from_target(self.input().get(k1))
                s2 = load_set_from_target(self.input().get(k2))
                output.write_tsv(k1, k2, len(s1), len(s2),
                                 len(s1.intersection(s2)))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class AIISSNStats(AITask):
    """ Match ISSN lists. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'crossref': CrossrefUniqISSNList(date=self.date),
            'degruyter': DegruyterISSNList(date=self.date),
            'doaj': DOAJISSNList(date=self.date),
            'jstor': JstorISSNList(date=self.date)
        }

    @timed
    def run(self):
        with self.output().open('w') as output:
            for k1, k2 in itertools.combinations(list(self.input().keys()), 2):
                s1 = load_set_from_target(self.input().get(k1))
                s2 = load_set_from_target(self.input().get(k2))
                output.write_tsv(k1, k2, len(s1), len(s2),
                                 len(s1.intersection(s2)))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class AIISSNOverlaps(AITask):
    """ Match ISSN lists. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'crossref': CrossrefUniqISSNList(date=self.date),
            'degruyter': DegruyterISSNList(date=self.date),
            'doaj': DOAJISSNList(date=self.date),
            'jstor': JstorISSNList(date=self.date)
        }

    @timed
    def run(self):
        with self.output().open('w') as output:
            for k1, k2 in itertools.combinations(list(self.input().keys()), 2):
                s1 = load_set_from_target(self.input().get(k1))
                s2 = load_set_from_target(self.input().get(k2))
                for issn in sorted(s1.intersection(s2)):
                    output.write_tsv(k1, k2, issn)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class AIISSNList(AITask):
    """
    List of uniq ISSNs in AI.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'crossref': CrossrefUniqISSNList(date=self.date),
            'degruyter': DegruyterISSNList(date=self.date),
            'doaj': DOAJISSNList(date=self.date),
            'jstor': JstorISSNList(date=self.date),
            'thieme': ThiemeISSNList(date=self.date),
            'elsevier': ElsevierJournalsISSNList(date=self.date),
        }

    @timed
    def run(self):
        _, output = tempfile.mkstemp(prefix='siskin-')
        for _, target in list(self.input().items()):
            shellout("cat {input} | grep -v null >> {output}",
                     input=target.path, output=output)
        output = shellout("sort -u {input} > {output}", input=output)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class AIQuality(AITask):
    """
    Create short quality report.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return [
            ArxivIntermediateSchema(date=self.date),
            CrossrefIntermediateSchema(date=self.date),
            DegruyterIntermediateSchema(date=self.date),
            DOAJIntermediateSchema(date=self.date),
            ElsevierJournalsIntermediateSchema(date=self.date),
            GeniosCombinedIntermediateSchema(date=self.date),
            IEEEIntermediateSchema(date=self.date),
            JstorIntermediateSchema(date=self.date),
            ThiemeIntermediateSchema(date=self.date),
        ]

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for target in self.input():
            shellout(""" echo "S:{input}" >> {output} """,
                     input=target.path, output=stopover)
            shellout(
                "span-check <(unpigz -c {input}) 2>&1 | jq . >> {output}", input=target.path, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='txt'))


class AIIntermediateSchema(AITask):
    """ Create an intermediate schema record from all AI sources. """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return [
            ArxivIntermediateSchema(date=self.date, stamp=True),
            CrossrefIntermediateSchema(date=self.date, stamp=True),
            DegruyterIntermediateSchema(date=self.date, stamp=True),
            DOAJIntermediateSchema(date=self.date, stamp=True),
            ElsevierJournalsIntermediateSchema(date=self.date, stamp=True),
            GenderopenIntermediateSchema(date=self.date, stamp=True),
            GeniosCombinedIntermediateSchema(date=self.date, stamp=True),
            IEEEIntermediateSchema(date=self.date, stamp=True),
            JstorIntermediateSchema(date=self.date, stamp=True),
            ThiemeIntermediateSchema(date=self.date, stamp=True),
            SpringerIntermediateSchema(stamp=True),
            PQDTIntermediateSchema(date=self.date, stamp=True),
            KielFMFIntermediateSchema(stamp=True),
            IJOCIntermediateSchema(stamp=True),
            CeeolJournalsDumpIntermediateSchema(stamp=True),
            SSOARIntermediateSchema(stamp=True),
            LyndaIntermediateSchema(date=self.date, stamp=True),
            HHBDIntermediateSchema(date=self.date, stamp=True),
        ]

    @timed
    def run(self):
        """
        Check, if all files are gzipped before we concatenate.
        """
        for target in self.input():
            with open(target.path) as f:
                if binascii.hexlify(f.read(2)) != b'1f8b':
                    raise RuntimeError('AIIntermediateSchema requires gzipped inputs, failed: %s' % target.path)

        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for target in self.input():
            shellout("cat {input} >> {output}",
                     input=target.path, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class AIAddReferences(AITask):
    """
    Experimental enrichment with linked documents.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'ai': AIIntermediateSchema(date=self.date),
            'db': MAGReferenceDB(date=self.date),
        }

    @timed
    def run(self):
        """
        Run query for DOI for each document and add a new keys if we find referenced DOI.
        """
        with sqlitedb(self.input().get('db').path) as cursor:
            with self.input().get('ai').open() as handle:
                with self.output().open('w') as output:
                    stmt = """
                    select doi from lookup where id IN (
                        select ref from refs where id = (select id from lookup where doi = ? ))
                    """
                    for line in handle:
                        doc = json.loads(line)
                        doi = doc.get('doi')
                        if doi is not None:
                            cursor.execute(stmt, (doi,))
                            rows = cursor.fetchall()
                            doc["refs"] = [v[0] for v in rows]
                        output.write(json.dumps(doc))
                        output.write("\n")

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class AICheckStats(AITask):
    """
    Run quality check and gather error records.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return AIIntermediateSchema(date=self.date)

    @timed
    def run(self):
        output = shellout(
            "span-check -verbose <(unpigz -c {input}) | pigz -c > {output}", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class AIErrorDistribution(AITask):
    """
    Error distribution.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return AICheckStats(date=self.date)

    @timed
    def run(self):
        output = shellout(
            "unpigz -c {input} | jq -rc .err | sort | uniq -c | sort -nr > {output}", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='txt'))


class AIRedact(AITask):
    """
    Redact intermediate schema.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return AIApplyOpenAccessFlag(date=self.date)

    @timed
    def run(self):
        """ A bit slower: `jq 'del(.["x.fulltext"])' input > output` """
        output = shellout(
            "span-redact <(unpigz -c {input}) | pigz -c > {output}", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class AIBlobDB(AITask):
    """
    Create data.db file for microblob, then:

    $ microblob -db $(taskoutput AIBlobDB) -file $(taskoutput AIRedact) -serve
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return AIRedact(date=self.date)

    @timed
    def run(self):
        """
        Extract intermediate schema file temporarily.
        """
        tempdir = tempfile.mkdtemp(prefix="siskin-")
        extracted = shellout(
            "unpigz -c {input} > {output}", input=self.input().path)

        shellout("microblob -db {tempdir} -file {input} -key finc.id",
                 tempdir=tempdir, input=extracted)
        os.remove(extracted)
        os.makedirs(os.path.dirname(self.output().path))
        shutil.move(tempdir, self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldb'))


class AILicensing(AITask):
    """
    Take intermediate schema and a config and attach ISILs accordingly. As per
    MDM-2017-11-29 a fixed date should be used. We fix date to YYYY-MM-15, refs
    #11821, #12021. AMSLFilterConfigFreeze should be run via a daily cron.

    Example crontab (assuming virtual environment named siskin):

        00 12  * * * source $HOME/.virtualenvs/siskin/bin/activate && taskdo AMSLFilterConfigFreeze --local-scheduler
    """
    date = ClosestDateParameter(default=datetime.date.today())
    override = luigi.BoolParameter(description="do not use jour fixe", significant=False)

    def requires(self):
        if self.override:
            jourfixe = self.date
        else:
            jourfixe = datetime.date(self.date.year, self.date.month, 15)
            if self.date.day < 15:
                jourfixe = jourfixe + relativedelta(months=-1)

        self.logger.debug("AILicensing date: %s (override=%s)", jourfixe, self.override)

        return {
            'is': AIApplyOpenAccessFlag(date=self.date),
            'config': AMSLFilterConfigFreeze(date=jourfixe),
        }

    def run(self):
        """
        span v0.1.204 or later.
        """
        output = shellout("span-tag -unfreeze {config} <(unpigz -c {input}) | pigz -c > {output}",
                          config=self.input().get('config').path, input=self.input().get('is').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class AILocalData(AITask):
    """
    Extract a CSV about source, id, doi and institutions for deduplication.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    batchsize = luigi.IntParameter(default=25000, significant=False)

    def requires(self):
        return AILicensing(date=self.date)

    def run(self):
        """
        Unzip on the fly, extract fields as CSV, sort be third column.
        """
        output = shellout("""unpigz -c {input} | span-local-data -b {size} | LC_ALL=C sort -S20% -t, -k3 > {output} """,
                          size=self.batchsize, input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='csv'))


class AIInstitutionChanges(AITask):
    """
    Calculate institution changes based on DOI duplicates. Experimental, using
    https://github.com/miku/groupcover with preferences.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return AILocalData(date=self.date)

    def run(self):
        output = shellout(
            """groupcover -lower -prefs '85 55 89 60 50 105 34 101 53 49 28 48 121' < {input} > {output}""", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class AIIntermediateSchemaDeduplicated(AITask):
    """
    A DOI deduplicated version of the intermediate schema. Experimental.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'changes': AIInstitutionChanges(date=self.date),
            'file': AILicensing(date=self.date),
        }

    def run(self):
        """
        Update intermediate schema labels from file. We cut out the ID and the
        ISIL list from the changes.
        """
        output = shellout("unpigz -c {input} | span-update-labels -b 20000 -f <(cut -d, -f1,4- {file}) | pigz -c > {output}",
                          input=self.input().get('file').path,
                          file=self.input().get('changes').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class AIExport(AITask):
    """
    Export to various formats

    XXX: #11467, crossref vs JSTOR.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='solr5vu3', description='export format')

    def requires(self):
        return AIIntermediateSchemaDeduplicated(date=self.date)

    def run(self):
        output = shellout("span-export -o {format} <(unpigz -c {input}) | pigz -c > {output}",
                          format=self.format, input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        extensions = {
            'solr5vu3': 'ldj.gz',
            'formeta': 'form.gz',
        }
        return luigi.LocalTarget(path=self.path(ext=extensions.get(self.format, 'gz')))


class AIUpdate(AITask, luigi.WrapperTask):
    """
    Just a wrapper task for updates, refs #5702.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return [AIExport(date=self.date), AIRedact(date=self.date)]

    def output(self):
        return self.input()


class AICollectionsAndSerialNumbers(AITask):
    """
    Turtlized coverage report, ..., refs #5156.

    XXX: The collection names are not that uniform. Would need to extract those
    names from AMSL - however, the raw data contains these names ...
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return AIIntermediateSchema(date=self.date)

    def run(self):
        """
        Create a graph with two namespaces.
        """
        g = rdflib.Graph()
        ns = {
            "amsl": rdflib.Namespace('http://amsl.technology/'),
            "disco": rdflib.Namespace('http://amsl.technology/discovery/Metadatenkollektion/'),
        }

        p = ns["amsl"].coveredMediumID
        disco = ns["disco"]

        with self.input().open() as handle:
            for i, line in enumerate(handle):
                if i % 100000 == 0:
                    self.logger.debug("%s %s", i, len(g))

                doc = json.loads(line)
                issns = list(itertools.chain(
                    doc.get('rft.issn', []), doc.get('rft.eissn', [])))
                colls = doc.get('finc.mega_collection', [])

                for issn in issns:
                    for c in colls:
                        s = disco[urllib.quote(c.encode('utf-8'))]

                        uri = 'urn:ISSN:%s' % urllib.quote(issn)
                        o = rdflib.URIRef(uri.strip())

                        g.add((s, p, o))

        with self.output().open('w') as output:
            output.write(g.serialize(format='turtle'))

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ttl'))


class AICoverageISSN(AITask):
    """
    For all AI sources and all holding files, find the number
    (perc) of ISSNs in the holding file, that we have some reach to.

    Result looks something like this, indicating availablity and sources:

    2091-2145   crossref|doaj
    2091-2234   crossref
    2091-2560   crossref
    2091-2609   crossref
    2091-2730   NOT_FOUND

    """
    date = ClosestDateParameter(default=datetime.date.today())
    isil = luigi.Parameter(default='DE-15')

    def requires(self):
        """
        TODO: Add GBI ISSN list.
        """
        return {
            'crossref': CrossrefUniqISSNList(date=self.date),
            'jstor': JstorISSNList(date=self.date),
            'degruyter': DegruyterISSNList(date=self.date),
            'doaj': DOAJISSNList(date=self.date),
            'gbi': GeniosISSNList(date=self.date),
            'elsevierjournals': ElsevierJournalsISSNList(date=self.date),
            'thieme': ThiemeISSNList(date=self.date),
            'file': AMSLHoldingsFile(isil=self.isil),
        }

    def run(self):
        issns = collections.defaultdict(set)

        with self.input().get('file').open() as handle:
            for row in handle:
                fields = row.strip().split('\t')
                if len(fields) < 3:
                    continue

                if re.search(r'[0-9]{4}-[0-9]{3}[0-9X]', fields[1]):
                    issns['file'].add(fields[1])

                if re.search(r'[0-9]{4}-[0-9]{3}[0-9X]', fields[2]):
                    issns['file'].add(fields[2])

        sources = ['crossref', 'jstor', 'degruyter',
                   'doaj', 'gbi', 'elsevierjournals', 'thieme']

        for source in sources:
            with self.input().get(source).open() as handle:
                for row in handle:
                    issns[source].add(row.strip())

        with self.output().open('w') as output:
            for issn in sorted(issns['file']):
                fields = []
                for source in sources:
                    if issn in issns[source]:
                        fields.append(source)
                if len(fields) == 0:
                    output.write_tsv(issn, "NOT_FOUND")
                else:
                    output.write_tsv(issn, "|".join(fields))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class AIISSNCoverageCatalogMatches(AITask):
    """
    For all ISSN, that are NOT_FOUND in AI, do a HTTP request to the current
    catalog and scrape the number of results.

    Example output:

    0171-4880   FOUND_RESULTS_1 https://katalog.ub.uni-...
    0171-5801   ERR_NOT_IN_CATALOG  https://katalog.ub....
    0171-5860   FOUND_RESULTS_466   https://katalog.ub....
    0171-7227   FOUND_RESULTS_2 https://katalog.ub.uni-...

    """
    date = ClosestDateParameter(default=datetime.date.today())
    isil = luigi.Parameter(default='DE-15')

    def requires(self):
        return AICoverageISSN(date=self.date, isil=self.isil)

    def run(self):
        if not self.isil == 'DE-15':
            raise RuntimeError('not implemented except for DE-15')

        cache = URLCache(directory=os.path.join(
            tempfile.gettempdir(), '.urlcache'))
        adapter = requests.adapters.HTTPAdapter(max_retries=3)
        cache.sess.mount('http://', adapter)

        with self.input().open() as handle:
            with self.output().open('w') as output:
                for i, row in enumerate(handle.iter_tsv(cols=('issn', 'status'))):
                    if row.status == 'NOT_FOUND':
                        link = 'https://katalog.ub.uni-leipzig.de/Search/Results?lookfor=%s&type=ISN' % row.issn
                        self.logger.info('fetch #%05d: %s' % (i, link))
                        body = cache.get(link)
                        if 'Keine Ergebnisse!' in body:
                            output.write_tsv(
                                row.issn, 'ERR_NOT_IN_CATALOG', link)
                        else:
                            soup = BeautifulSoup(body)
                            rs = soup.findAll("div", {"class": "floatleft"})
                            if len(rs) == 0:
                                output.write_tsv(row.issn, 'ERR_LAYOUT', link)
                                continue
                            first = rs[0]
                            match = re.search(
                                r'Treffer([0-9]+)-([0-9]+)von([0-9]+)', first.text)
                            if match:
                                total = match.group(3)
                                output.write_tsv(
                                    row.issn, 'FOUND_RESULTS_%s' % total, link)
                            else:
                                output.write_tsv(
                                    row.issn, 'ERR_NO_MATCH', link)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class AIISSNCoverageSolrMatches(AITask):
    date = ClosestDateParameter(default=datetime.date.today())
    isil = luigi.Parameter(default='DE-15')

    def requires(self):
        return AICoverageISSN(date=self.date, isil=self.isil)

    def run(self):
        if self.isil != 'DE-15':
            raise RuntimeError('not implemented except for DE-15')

        cache = URLCache(directory=os.path.join(
            tempfile.gettempdir(), '.urlcache'))
        adapter = requests.adapters.HTTPAdapter(max_retries=3)
        cache.sess.mount('http://', adapter)

        finc = self.config.get('ai', 'finc-solr')
        ai = self.config.get('ai', 'ai-solr')

        def numFound(link):
            """ Given a SOLR query URL, return the number of docs found. """
            body = cache.get(link)
            content = json.loads(body)
            return content['response']['numFound']

        with self.input().open() as handle:
            with self.output().open('w') as output:
                for i, row in enumerate(handle.iter_tsv(cols=('issn', 'status'))):
                    if row.status == 'NOT_FOUND':
                        link = '%s/select?q=institution:%s+AND+issn:%s&wt=json' % (
                            finc, self.isil, row.issn)
                        self.logger.info('fetch #%05d: %s', i, link)
                        output.write_tsv('finc', row.issn,
                                         numFound(link), link)
                    else:
                        link = '%s/select?q=institution:%s+AND+issn:%s&wt=json' % (
                            ai, self.isil, row.issn)
                        self.logger.info('fetch #%05d: %s', i, link)
                        output.write_tsv('ai', row.issn, numFound(link), link)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class AIApplyOpenAccessFlag(AITask):
    """
    Apply OA-Flag. Experimental, refs. #8986.

    1. Inhouse/Flag (per collection)
    2. KBART
    3. Gold-OA
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'amslfc': AMSLFreeContent(date=self.date),
            'kbart': AMSLOpenAccessKBART(date=self.date),
            'file': AIIntermediateSchema(date=self.date),
        }

    def run(self):
        """
        Python: 500k recs/min, Go (span-oa-filter): 2.5M recs/min, refs #11285#note-17, refs #11969.

        XXX: Testing: Adjust filtered file with data from AMSLFreeContent.

        Exclude 48 explicitlty (span 0.1.225 and later) with -xsid, refs #12738.
        """

        output = shellout("""unpigz -c {input} |
                             span-oa-filter -xsid 48 -f {kbart} -fc {amslfc} |
                             pigz -c > {output}""",
                          input=self.input().get('file').path,
                          kbart=self.input().get('kbart').path,
                          amslfc=self.input().get('amslfc').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Peek [94].

[ddnl]

ftp-host = sftp://example.com
ftp-base = /target_data/xyz
ftp-username= admin
ftp-password= s3cr3t
ftp-pattern = *
"""

import datetime

import luigi
from gluish.format import TSV
from gluish.parameter import ClosestDateParameter

from siskin.common import FTPMirror
from siskin.task import DefaultTask

class DDNLTask(DefaultTask):
    """
    Base NL via DD.
    """
    TAG = 'nl'

class DDNLPaths(DDNLTask):
    """
    Mirror.
    """

    date = ClosestDateParameter(default=datetime.date.today())
    max_retries = luigi.IntParameter(default=10, significant=False)
    timeout = luigi.IntParameter(
        default=20, significant=False, description='timeout in seconds')

    def requires(self):
        return FTPMirror(host=self.config.get('ddnl', 'ftp-host'),
                         base=self.config.get('ddnl', 'ftp-base'),
                         username=self.config.get('ddnl', 'ftp-username'),
                         password=self.config.get('ddnl', 'ftp-password'),
                         pattern=self.config.get('ddnl', 'ftp-pattern'),
                         max_retries=self.max_retries,
                         timeout=self.timeout)

    def run(self):
        self.input().move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



#!/usr/bin/env python
# coding: utf-8
# pylint: disable=C0103

"""
Convert XML to JSON, standalone script.
"""

import json
import sys

import xmltodict

if __name__ == '__main__':
    if len(sys.argv) == 3:
        with open(sys.argv[2], 'w') as output:
            with open(sys.argv[1], 'r') as handle:
                doc = xmltodict.parse(handle.read())
                json.dump(doc, output)
    else:
        print("%s %s %s" % (sys.argv[0], 'input.xml', 'output.json'))



# coding: utf-8
# pylint: disable=C0103,W0232,C0301,W0703

# Copyright 2018 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
A new module for conversions. Complement assets.
"""

import base64
import collections
import logging
import tarfile
import tempfile
from xml.sax.saxutils import escape, unescape

import xmltodict

import marcx
import pymarc

html_escape_table = {
    '"': "&quot;",
    "'": "&apos;"
}
html_unescape_table = {v: k for k, v in html_escape_table.items()}

marburg_language_mapping = {
    "de": "ger",
    "deu": "ger",
}

logger = logging.getLogger('siskin')

def html_escape(text):
    """
    Escape HTML, see also: https://wiki.python.org/moin/EscapingHtml
    """
    return escape(text, html_escape_table)

def html_unescape(text):
    """
    Unescape HTML, see also: https://wiki.python.org/moin/EscapingHtml
    """
    return unescape(text, html_unescape_table)

def imslp_tarball_to_marc(tarball, outputfile=None, legacy_mapping=None,
                          max_failures=30):
    """
    Convert an IMSLP tarball to MARC binary output file without extracting it.
    If outputfile is not given, write to a temporary location.

    Returns the location of the resulting MARC file.

    A maximum number of failed conversions can be specified with `max_failures`,
    as of 2018-04-25, there were 30 records w/o title.
    """
    if outputfile is None:
        _, outputfile = tempfile.mkstemp(prefix="siskin-")

    stats = collections.Counter()

    with open(outputfile, "wb") as output:
        writer = pymarc.MARCWriter(output)
        with tarfile.open(tarball) as tar:
            for member in tar.getmembers():
                fobj = tar.extractfile(member)
                try:
                    record = imslp_xml_to_marc(fobj.read(), legacy_mapping=legacy_mapping)
                    writer.write(record)
                except ValueError as exc:
                    logger.warn("conversion failed: %s", exc)
                    stats["failed"] += 1
                finally:
                    fobj.close()
                    stats["processed"] += 1

        writer.close()

        if stats["failed"] > max_failures:
            raise RuntimeError("more than %d records failed", max_failures)

        logger.debug("%d/%d records failed/processed", stats["failed"], stats["processed"])

    return outputfile

def imslp_xml_to_marc(s, legacy_mapping=None):
    """
    Convert a string containing a single IMSLP XML record to a pymarc MARC record.

    Optionally take a legacy mapping, associating IMSLP Names with VIAF
    identifiers. Blueprint: https://git.io/vpQPd, with one difference: We
    allow records with no subjects.

    A record w/o title is an error. We check for them, when we add the field.
    """
    dd = xmltodict.parse(s, force_list={"subject", "languages"})

    if legacy_mapping is None:
        legacy_mapping = collections.defaultdict(lambda: collections.defaultdict(str))

    record = marcx.Record(force_utf8=True)
    record.strict = False

    doc = dd["document"]

    record.leader = "     ncs  22        450 "

    identifier = doc["identifier"]["#text"]
    encoded_id = base64.b64encode(identifier.encode("utf-8")).rstrip("=")
    record.add("001", data=u"finc-15-{}".format(encoded_id))

    record.add("007", data="cr")

    if doc.get("languages", []):
        langs = [l for l in doc["languages"] if l != "unbekannt"]
        if langs:
            record.add("008", data="130227uu20uuuuuuxx uuup%s  c" % langs[0])
            for l in langs:
                record.add("041", a=l)

    creator = doc["creator"]["mainForm"]
    record.add("100", a=creator, e="cmp",
               _0=legacy_mapping.get(identifier, {}).get("viaf", ""))

    record.add("240", a=legacy_mapping.get(identifier, {}).get("title", ""))

    try:
        record.add("245", a=html_unescape(doc["title"]))
    except KeyError:
        raise ValueError("cannot find title: %s ..." % s[:300])

    record.add("246", a=html_unescape(doc.get("additionalTitle", "")))

    record.add("260", c=doc.get("date", ""))
    record.add("650", y=doc.get("date", ""))
    record.add("500", a=doc.get("abstract", ""))

    for689 = []

    if "subject" in doc:
        if len(doc["subject"]) == 1:
            for689.append(doc["subject"][0]["mainForm"])
        elif len(doc["subject"]) == 2:
            for689.append(doc["subject"][1]["mainForm"])
        else:
            raise ValueError("cannot handle %d subjects", len(doc["subject"]))

        record.add("590", a=for689[0].title(),
                   b=doc.get("music_arrangement_of", "").title())

        for689.append(doc.get("music_arrangement_of", ""))

        for subject in set(for689):
            record.add("689", a=subject.title())

    record.add("700", a=doc.get("contributor", {}).get("mainForm", ""), e="ctb")
    record.add("856", q="text/html", _3="Petrucci Musikbibliothek",
               u=doc["url"]["#text"])
    record.add("970", c="PN")
    record.add("980", a=identifier, b="15", c="Petrucci Musikbibliothek")
    return record

def marburg_to_marc(s):
    """
    Convert a string containing a single XML in datacite from
    http://archiv.ub.uni-marburg.de/ubfind/OAI/Server into a binary MARC.
    """
    dd = xmltodict.parse(s, force_list={"dcite:creators", "dcite:titles"})
    record = marcx.Record()

    header, metadata = dd["Record"]["header"], dd["Record"]["metadata"]
    identifier = header["identifier"]

    record.add("001", data=base64.b64encode(identifier).rstrip("="))
    record.add("007", data="cr")

    related = metadata["dcite:resource"]["dcite:relatedIdentifiers"]
    for rid in related:
        if rid["@relatedIdentifierType"] != "ISSN":
            continue
        record.add("022", a=rid["#text"])

    language = metadata["dcite:resource"]["dcite:language"]
    language = marburg_language_mapping.get(language)
    record.add("008", data="130227uu20uuuuuuxx uuup%s  c" % language)
    record.add("041", a=marburg_language_mapping.get(language))

    for item in metadata["dcite:resource"]["dcite:creators"]:
        name = item["dcite:creator"]["dcite:creatorName"]["#text"]
        record.add("100", a=name)

    for item in metadata["dcite:resource"]["dcite:titles"]:
        title = item["dcite:title"]

    # WIP.

# Facet fields values from author2_role field across all sources.
# TODO: fill out empty strings, maybe add https://git.io/fNLwY, too.
author_role_mapping = {
    "\xc3\xbcbers": "",
    "\xc3\xbcbersetzer": "",
    "\xc3\xbcbersetzerin": "",
    "arrangeur": "",
    "bearb": "",
    "clb": "",
    "darst": "",
    "donateur": "",
    "ed": "",
    "edtomas": "",
    "fotografin": "",
    "herausgeber": "",
    "herausgeberin": "",
    "hg": "",
    "hrsg": "",
    "illustrator": "",
    "kommentatorin": "",
    "komponist": "",
    "librettist": "",
    "mitarb": "",
    "mitarbeiter": "",
    "mitwirkender": "",
    "photographer": "",
    "pord": "",
    "sonstige": "",
    "t\xc3\xa4nzer": "",
    "textdichter": "",
    "textverf": "",
    "veranstalter": "",
    "verfasser einer einleitung": "",
    "verfasser eines nachworts": "",
    "verfasser eines vorworts": "",
    "verfasser von kommentaren": "",
    "verfasser": "",
    "verfasserin einer einleitung": "",
    "verfasserin eines geleitwortes": "",
    "verfasserin eines nachworts": "",
    "verfasserin eines vorworts": "",
    "verfasserin von erg\xc3\xa4nzendem text": "",
    "verfasserin von zusatztexten": "",
    "verfasserin": "",
    "voc": "",
    "zusammenstellender": "",

    # More, from MAB/151 mostly.
    " [Animation]": "",
    " [Ausstattung]": "",
    " [Darst.]": "",
    " [Drehbuch]": "",
    " [Kamera]": "",
    " [Komp.]": "",
    " [Musik]": "",
    " [Prod.]": "",
    " [Regie]": "",
    " [Schnitt]": "",
    " [Ton]": "",
    ", [Drehbuch]": "",
    "Darst.]": "",
    "Drehbuch]": "",
    "Prod.]": "",
    "[Alt]": "",
    "[Animation]": "",
    "[Assistent]": "",
    "[Assistentin]": "",
    "[Aufnahmeleiter]": "",
    "[Ausstattung]": "",
    "[Autor]": "",
    "[Bariton]": "",
    "[Bass]": "",
    "[Bearb.]": "",
    "[Begr.]": "",
    "[Beleuchtung]": "",
    "[Beratung]": "",
    "[Buch]": "",
    "[Casting]": "",
    "[Co-Produzent]": "",
    "[Computeranimation]": "",
    "[DRehbuch]": "",
    "[Darst.]": "",
    "[Darst.r]": "",
    "[Darst]": "",
    "[Darsteller]": "",
    "[Dehbuch]": "",
    "[Digitale Bildgestaltung]": "",
    "[Dir.]": "",
    "[Donat]": "",
    "[Donateur]": "",
    "[Doz.]": "",
    "[Dozent]": "",
    "[Dozentin]": "",
    "[Dramaturgie]": "",
    "[Drast.]": "",
    "[Drebuch]": "",
    "[Drehbuch]": "",
    "[Flöte]": "",
    "[Fotogr.]": "",
    "[Gitarre]": "",
    "[Herst.-Leitung]": "",
    "[Hrsg.]": "",
    "[Idee]": "",
    "[Ill.]": "",
    "[Inszenierung]": "",
    "[Interpret]": "",
    "[Interviewer]": "",
    "[Kamera]": "",
    "[Klavier]": "",
    "[Komp.]": "",
    "[Kostüm]": "",
    "[Kostüme]": "",
    "[Leitung]": "",
    "[Licht]": "",
    "[Malerei]": "",
    "[Mezzosopran]": "",
    "[Mitarb.]": "",
    "[Mitverf.]": "",
    "[Mitwirkende]": "",
    "[Musik]": "",
    "[Nachr.]": "",
    "[Orgel]": "",
    "[Originalkonzept]": "",
    "[Originalvorl.]": "",
    "[Originalvorlage]": "",
    "[Prod.-Leitung]": "",
    "[Prod.]": "",
    "[Prod]": "",
    "[Produzent]": "",
    "[Projektbetreuer]": "",
    "[Projektbetreuerin]": "",
    "[Recherche]": "",
    "[Red.]": "",
    "[Redaktion SWR]": "",
    "[Redaktion]": "",
    "[Reg.]": "",
    "[Regie]": "",
    "[Saxophon]": "",
    "[Schnitt]": "",
    "[Schöpfer]": "",
    "[Sopran]": "",
    "[Special-Effects]": "",
    "[Spezialeffekte]": "",
    "[Sprecher]": "",
    "[Sprecherin]": "",
    "[Studiengangkoordinator]": "",
    "[Studiengangkoordinatorin]": "",
    "[Studiengangskoordinator]": "",
    "[Studiengangskoordinatorin]": "",
    "[Szenebild]": "",
    "[Szenenbild]": "",
    "[Szenerie]": "",
    "[Szenographie]": "",
    "[Tenor]": "",
    "[Ton]": "",
    "[Tonschnitt]": "",
    "[Violine]": "",
    "[Violoncello]": "",
    "[Visual Effects]": "",
    "[Vorl.]": "",
    "[Vorl]": "",
    "[Vorlage]": "",
    "[Vorr.]": "",
    "[Zeichner]": "",
    "[darst.]": "",
    "[gefeierte Pers.]": "",
    "[guitar]": "",
    "[schnitt]": "",
    "[vocal]": "",
    "[Übers.]": "",

    # Below terms are actually ok, as per:
    # http://www.loc.gov/marc/relators/relaterm.html
    "abr": "abr",
    "acp": "acp",
    "act": "act",
    "adi": "adi",
    "adp": "adp",
    "aft": "aft",
    "anl": "anl",
    "anm": "anm",
    "ann": "ann",
    "ant": "ant",
    "ape": "ape",
    "apl": "apl",
    "app": "app",
    "aqt": "aqt",
    "arc": "arc",
    "ard": "ard",
    "arr": "arr",
    "art": "art",
    "asg": "asg",
    "asn": "asn",
    "ato": "ato",
    "att": "att",
    "auc": "auc",
    "aud": "aud",
    "aui": "aui",
    "aus": "aus",
    "aut": "aut",
    "bdd": "bdd",
    "bjd": "bjd",
    "bkd": "bkd",
    "bkp": "bkp",
    "blw": "blw",
    "bnd": "bnd",
    "bpd": "bpd",
    "brd": "brd",
    "brl": "brl",
    "bsl": "bsl",
    "cas": "cas",
    "ccp": "ccp",
    "chr": "chr",
    "cli": "cli",
    "cll": "cll",
    "clr": "clr",
    "clt": "clt",
    "cmm": "cmm",
    "cmp": "cmp",
    "cmt": "cmt",
    "cnd": "cnd",
    "cng": "cng",
    "cns": "cns",
    "coe": "coe",
    "col": "col",
    "com": "com",
    "con": "con",
    "cor": "cor",
    "cos": "cos",
    "cot": "cot",
    "cou": "cou",
    "cov": "cov",
    "cpc": "cpc",
    "cpe": "cpe",
    "cph": "cph",
    "cpl": "cpl",
    "cpt": "cpt",
    "cre": "cre",
    "crp": "crp",
    "crr": "crr",
    "crt": "crt",
    "csl": "csl",
    "csp": "csp",
    "cst": "cst",
    "ctb": "ctb",
    "cte": "cte",
    "ctg": "ctg",
    "ctr": "ctr",
    "cts": "cts",
    "ctt": "ctt",
    "cur": "cur",
    "cwt": "cwt",
    "dbp": "dbp",
    "dfd": "dfd",
    "dfe": "dfe",
    "dft": "dft",
    "dgg": "dgg",
    "dgs": "dgs",
    "dis": "dis",
    "dln": "dln",
    "dnc": "dnc",
    "dnr": "dnr",
    "dpc": "dpc",
    "dpt": "dpt",
    "drm": "drm",
    "drt": "drt",
    "dsr": "dsr",
    "dst": "dst",
    "dtc": "dtc",
    "dte": "dte",
    "dtm": "dtm",
    "dto": "dto",
    "dub": "dub",
    "edc": "edc",
    "edm": "edm",
    "edt": "edt",
    "egr": "egr",
    "elg": "elg",
    "elt": "elt",
    "eng": "eng",
    "enj": "enj",
    "etr": "etr",
    "evp": "evp",
    "exp": "exp",
    "fac": "fac",
    "fds": "fds",
    "fld": "fld",
    "flm": "flm",
    "fmd": "fmd",
    "fmk": "fmk",
    "fmo": "fmo",
    "fmp": "fmp",
    "fnd": "fnd",
    "fpy": "fpy",
    "frg": "frg",
    "gis": "gis",
    "grt": "grt",
    "his": "his",
    "hnr": "hnr",
    "hst": "hst",
    "ill": "ill",
    "ilu": "ilu",
    "ins": "ins",
    "inv": "inv",
    "isb": "isb",
    "itr": "itr",
    "ive": "ive",
    "ivr": "ivr",
    "jud": "jud",
    "jug": "jug",
    "lbr": "lbr",
    "lbt": "lbt",
    "ldr": "ldr",
    "led": "led",
    "lee": "lee",
    "lel": "lel",
    "len": "len",
    "let": "let",
    "lgd": "lgd",
    "lie": "lie",
    "lil": "lil",
    "lit": "lit",
    "lsa": "lsa",
    "lse": "lse",
    "lso": "lso",
    "ltg": "ltg",
    "lyr": "lyr",
    "mcp": "mcp",
    "mdc": "mdc",
    "med": "med",
    "mfp": "mfp",
    "mfr": "mfr",
    "mod": "mod",
    "mon": "mon",
    "mrb": "mrb",
    "mrk": "mrk",
    "msd": "msd",
    "mte": "mte",
    "mtk": "mtk",
    "mus": "mus",
    "nrt": "nrt",
    "opn": "opn",
    "org": "org",
    "orm": "orm",
    "osp": "osp",
    "oth": "oth",
    "own": "own",
    "pan": "pan",
    "pat": "pat",
    "pbd": "pbd",
    "pbl": "pbl",
    "pdr": "pdr",
    "pfr": "pfr",
    "pht": "pht",
    "plt": "plt",
    "pma": "pma",
    "pmn": "pmn",
    "pop": "pop",
    "ppm": "ppm",
    "ppt": "ppt",
    "pra": "pra",
    "prc": "prc",
    "prd": "prd",
    "pre": "pre",
    "prf": "prf",
    "prg": "prg",
    "prm": "prm",
    "prn": "prn",
    "pro": "pro",
    "prp": "prp",
    "prs": "prs",
    "prt": "prt",
    "prv": "prv",
    "pta": "pta",
    "pte": "pte",
    "ptf": "ptf",
    "pth": "pth",
    "ptt": "ptt",
    "pup": "pup",
    "rbr": "rbr",
    "rcd": "rcd",
    "rce": "rce",
    "rcp": "rcp",
    "rdd": "rdd",
    "red": "red",
    "ren": "ren",
    "res": "res",
    "rev": "rev",
    "rpc": "rpc",
    "rps": "rps",
    "rpt": "rpt",
    "rpy": "rpy",
    "rse": "rse",
    "rsg": "rsg",
    "rsp": "rsp",
    "rsr": "rsr",
    "rst": "rst",
    "rth": "rth",
    "rtm": "rtm",
    "sad": "sad",
    "sce": "sce",
    "scl": "scl",
    "scr": "scr",
    "sds": "sds",
    "sec": "sec",
    "sgd": "sgd",
    "sgn": "sgn",
    "sht": "sht",
    "sll": "sll",
    "sng": "sng",
    "spk": "spk",
    "spn": "spn",
    "spy": "spy",
    "srv": "srv",
    "std": "std",
    "stg": "stg",
    "stl": "stl",
    "stm": "stm",
    "stn": "stn",
    "str": "str",
    "tcd": "tcd",
    "tch": "tch",
    "ths": "ths",
    "tld": "tld",
    "tlp": "tlp",
    "trc": "trc",
    "trl": "trl",
    "tyd": "tyd",
    "tyg": "tyg",
    "uvp": "uvp",
    "vac": "vac",
    "vdg": "vdg",
    "wac": "wac",
    "wal": "wal",
    "wam": "wam",
    "wat": "wat",
    "wdc": "wdc",
    "wde": "wde",
    "win": "win",
    "wit": "wit",
    "wpr": "wpr",
    "wst": "wst",
}




# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301
# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Tracy Hoffmann, <tracy.hoffmann@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
KHM Koeln, refs #9269, #8391.

Update usually comes in two tarballs, carrying a 8 digit date string, but file
location might differ.

Configuration keys:

[khm]

scp-src = user@ftp.example.com:/home/gbi

"""

import datetime
import itertools
import os
import re
import tempfile

import luigi
from gluish.common import Executable
from gluish.format import TSV
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout

from siskin.benchmark import timed
from siskin.common import FTPMirror
from siskin.task import DefaultTask
from siskin.utils import iterfiles

class KHMTask(DefaultTask):
    """ Base task for KHM Koeln (sid 109) """
    TAG = '109'
    FILEPATTERN = r'.*aleph.ALL_RECS.(?P<date>[12][0-9]{3,3}[01][0-9][0123][0-9]).*(?P<no>[0-9]{1,}).tar.gz'

    def closest(self):
        return monthly(date=self.date)

class KHMDropbox(KHMTask):
    """
    Pull down content from FTP.
    """
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return Executable('rsync', message='https://rsync.samba.org/')

    def run(self):
        target = os.path.join(self.taskdir(), 'mirror')
        shellout("mkdir -p {target} && rsync {rsync_options} {src} {target}",
                 rsync_options=self.config.get('khm', 'rsync-options', fallback='-avzP'),
                 src=self.config.get('khm', 'scp-src'), target=target)

        if not os.path.exists(self.taskdir()):
            os.makedirs(self.taskdir())

        with self.output().open('w') as output:
            for path in sorted(iterfiles(target)):
                output.write_tsv(path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist'), format=TSV)

class KHMLatest(KHMTask):
    """
    Write out a file with latest date and path to latest files.

    20180725        .../109/KHMDropbox/mirror/khmkoeln/august-2018/aleph.ALL_RECS.20180725.125558.2.tar.gz
    20180725        .../109/KHMDropbox/mirror/khmkoeln/august-2018/aleph.ALL_RECS.20180725.125255.1.tar.gz
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return KHMDropbox(date=self.date)

    def run(self):
        """
        Naming schema: aleph.ALL_RECS.20170316.123655.2.tar.gz. aleph.ALL_RECS.YYYYMMDD.HHMMSS.NO.tar.gz
        """
        with self.input().open() as handle:
            with self.output().open('w') as output:
                records = sorted(handle.iter_tsv(cols=('path',)), reverse=True,
                                 key=lambda row: os.path.basename(row.path))
                if len(records) == 0:
                    raise RuntimeError('no files found, cannot determine latest date')
                groups = re.search(self.FILEPATTERN, records[0].path).groupdict()
                latest_date = groups['date']
                for record in records:
                    gg = re.search(self.FILEPATTERN, record.path).groupdict()
                    if gg['date'] == latest_date:
                        output.write_tsv(latest_date, record.path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class KHMMARC(KHMTask):
    """
    Convert tarball to binary MARC. Note: we expect exactly two tarballs, which
    is a bit brittle at the moment.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return KHMLatest(date=self.date)

    def run(self):
        tarballs = []
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('x', 'path')):
                tarballs.append(row.path)

        if not len(tarballs) == 2:
            raise RuntimeError('limitation: we need exactly two tarballs')

        output = shellout("python {script} {t1} {t2} {output}",
                          script=self.assets("109/109_marcbinary.py"),
                          t1=tarballs[0], t2=tarballs[1])
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="fincmarc.mrc"))



#!/usr/bin/dev python3
# coding: utf-8

from __future__ import print_function

import base64
import io
import re
import sys

import marcx


def get_field(field):

    if field.find("</") != -1:
        pattern = field
    else:
        pattern = "<%s.*?>(.*?)<\/%s>" % (field, field)

    regexp = re.search(pattern, line)
    if regexp:
        value = regexp.group(1)
        return value
    else:
        return ""


def get_repeatable_field(field):

    if field.find("</") != -1:
        pattern = field
    else:
        pattern = "(<%s.*?>.*<\/%s>)" % (field, field)

    regexp = re.search(pattern, line)
    if regexp:
        value = regexp.group(1)
        value = value.replace("<%s>" % field, "")
        value = value.split("</%s>" % field)
    else:
        value = []
    return value

input, output = "39_input.xml", "39_output.mrc"

if len(sys.argv) == 3:
    input, output = sys.argv[1], sys.argv[2]

inputfile = io.open(input, "r", encoding="utf-8")
outputfile = io.open(output, "wb")

for i, line in enumerate(inputfile, start=1):

    line = line.replace("\n", "")

    # Zugangsrechte
    if "<dcterms:accessRights>restricted</dcterms:accessRights>" in line:
        continue

    # Identfier
    f001 = get_field("identifier")
    if f001 == "":
        continue
    f001 = f001
    f001 = f001.encode("utf-8")
    f001 = base64.b64encode(f001)
    f001 = f001.decode("ascii")
    f001 = f001.rstrip("=")

    # ISSN
    f022a = get_field("<dc:identifier scheme=\"ISSN\">issn:(.*)</dc:identifier>")

    # Sprache
    f041a = get_field("<dc:language>(.*)</dc:language>")

    # Person
    persons = get_repeatable_field("dc:creator")

    # Titel
    f245a = get_field("dc:title")
    try:
        f245a, f245b = f245a.split(" : ")
    except:
        f245b = ""

    # Erscheinungsvermerk
    f260b = get_field("dc:publisher")
    try:
        f260a, f260b = f260b.split(" : ")
    except:
        f260a = ""
    f260c = get_field("dc:date")

    # Umfang
    f300a = get_field("dcterms:extent")
    f300a = f300a
    if f300a != "":
        regexp = re.search("(\d+)\D", f300a)
        if regexp:
            f300a = regexp.group(1)
    else:
        f300a = ""

    # ausführliche Quellenangabe
    f500a = get_field("dcterms:bibliographicCitation")

    # beschreibender Text
    f520a = get_field("<dc:description xml:lang=\"\w\w\w\">(.*?)</dc:description>")

    FIELD_LENGTH_LIMIT = 8000  # Limit is 9999 bytes, but as we might not only have ASCII, add a margin.

    if len(f520a) > FIELD_LENGTH_LIMIT:
        print("cutting off long description (%s -> %s) for %s" % (len(f520a), FIELD_LENGTH_LIMIT, get_field("identifier")), file=sys.stderr)
        f520a = f520a[:FIELD_LENGTH_LIMIT]  # Cut off long abstracts, refs #11349.

    format = get_field("dc:type")
    if format == "notebiblio":
        leader = "     naa  22        4500"
    elif format == "illustration":
        leader = "     ckm  22        4500"
    elif format == "liminaire":
        leader = "     naa  22        4500"
    elif format == "table":
        leader = "     naa  22        4500" 
    elif format == "book":
        leader = "     cam  22        4500"   
    else:
        leader = "     cab  22        4500"

    # Quelle
    f773a = get_field("dcterms:bibliographicCitation\.jtitle")
    volume = get_field("dcterms:bibliographicCitation\.volume")
    issue = get_field("dcterms:bibliographicCitation\.issue")
    spage = get_field("dcterms:bibliographicCitation\.spage")
    epage = get_field("dcterms:bibliographicCitation\.epage")
    f773g = "%s(%s)%s, S. %s-%s" % (volume, f260c, issue, spage, epage)
    f773x = get_field("<dc:identifier scheme=\"ISSN\">issn:(.*)</dc:identifier>")

    # URL und DOI
    f856u = get_field("dc:identifier")
    doi = get_field("<dc:identifier scheme=\"DOI\">(.*?)</dc:identifier>")

    # Schlagwörter
    f950a = get_field("<dc:subject xml:lang=\"\w\w\w\">(.*?)</dc:subject>")
    f950a = f950a.split(" ; ")

    marcrecord = marcx.Record(force_utf8=True, to_unicode=True)
    marcrecord.strict = False
    marcrecord.leader = leader
    marcrecord.add("001", data="finc-39-%s" % f001)
    marcrecord.add("007", data="cr")
    #marcrecord.add("008", data=f008)
    marcrecord.add("022", a=f022a)
    marcrecord.add("041", a=f041a)

    if len(persons) > 0:
        f100a = persons[0]
    else:
        f100a = ""

    marcrecord.add("100", a=f100a)
    marcrecord.add("245", subfields=["a", f245a, "b", f245b])
    marcrecord.add("260", a=f260a, b=f260b, c=f260c)
    marcrecord.add("300", a=f300a)
    marcrecord.add("500", a=f500a)
    marcrecord.add("520", a=f520a)

    if len(persons) > 1:
        for f700a in persons[1:]:
            marcrecord.add("700", a=f700a)

    marcrecord.add("773", a=f773a, g=f773g, x=f773x)
    marcrecord.add("856", q="text/html", _3="Link zur Ressource", u=f856u)

    if doi != "":
        marcrecord.add("856", q="text/html", _3="DOI", u="http://doi.org/" + doi)

    for subject in f950a:
        marcrecord.add("950", a=subject)

    collections = ["a", f001, "b", "39", "c", u"Persée", "c", u"Persée (adlr)"]
    marcrecord.add("980", subfields=collections)

    outputfile.write(marcrecord.as_marc())

inputfile.close()
outputfile.close()



#!/usr/bin/env python
# coding: utf-8
# pylint: disable=C0103,C0301

"""
Custom conversion for IMSLP XML. Usually a set of files within a given
directory, refs #1240.

Usage:

    $ python 15_marcbinary.py [INPUT-DIRECTORY [, OUTPUT-FILE]]

"""

from __future__ import print_function

import io
import os
import re
import sys
import json
from xml.sax.saxutils import escape, unescape

import marcx
import xmltodict
import collections

langmap = {
    "Ancient Greek": "grc",
    "Armenian": "arm",
    "Catalan": "cat",
    "Catalàn": "cat",
    "Church Slavonic": "chu",
    "Croatian": "hrv",
    "Czech": "cze",
    "Danish": "dan",
    "Deutsch": "ger",
    "Dutch": "dut",
    "English": "eng",
    "Esperanto": "epo",
    "Finnish": "fin",
    "Français": "fre",
    "french": "fre",
    "French": "fre",
    "German": "ger",
    "german": "ger",
    "Greek": "ell",
    "Hebrew": "heb",
    "Hungarian": "hun",
    "Icelandic": "ice",
    "Irish": "gle",
    "Italian": "ita",
    "italian": "ita",
    "Italiano": "ita",
    "Italien": "ita",
    "Japanese": "jpn",
    "Latein": "lat",
    "Latin": "lat",
    "latin": "lat",
    "Middle French": "frm",
    "Neapolitan": "nap",
    "Norwegian": "nor",
    "Polish": "pol",
    "Portuguese": "por",
    "Romanian": "rom",
    "Russian": "rus",
    "Serbian (Cyrillic)": "srp",
    "Serbian": "srp",
    "Spanish": "spa",
    "Swedish": "swe",
    "Telugu": "tel",
    "Turkish": "tur",
    "Ukrainian": "ukr",
    "Welsh": "wel",
    "Yiddish": "yid",
}

html_escape_table = {
    '"': "&quot;",
    "'": "&apos;"
}
html_unescape_table = {v: k for k, v in html_escape_table.items()}


def html_escape(text):
    """ Escape HTML, see also: https://wiki.python.org/moin/EscapingHtml"""
    return escape(text, html_escape_table)


def html_unescape(text):
    """ Unescape HTML, see also: https://wiki.python.org/moin/EscapingHtml"""
    return unescape(text, html_unescape_table)

def get_field(tag):
    try:
        return record[tag]
    except:
        return ""


input_directory = "IMSLP_alt"

fieldmap = collections.defaultdict(dict)

for root, _, files in os.walk(input_directory):
    for filename in files:
        if not filename.endswith(".xml"):
            continue
        filepath = os.path.join(root, filename)
        inputfile = io.open(filepath, "r", encoding="utf-8")
       
        title = False
        f001 = ""       
        f1000 = ""
        f245a = ""

        for line in inputfile:
       
            regexp = re.search("<recordId>(.*)<\/recordId>", line)
            if regexp:
                f001 = regexp.group(1)
                fieldmap[f001]["title"] = f245a                
                continue

            # Deutscher Werktitel
            if title:
                regexp = re.search(r"<string>(.*)<\/string>", line)
                if regexp:                    
                    f245a = regexp.group(1)
                    f245a = html_unescape(f245a)                 

            regexp = re.search(r"name=\"worktitle\"", line)
            title = True if regexp else False            

            # VIAF
            regexp = re.search(r"<viafId>(\d+)</viafId>", line)  # Feld manchmal vorhanden, aber leer
            if regexp:
                f1000 = regexp.group(1)
                f1000 = "(VIAF)" + f1000
                fieldmap[f001]["viaf"] = f1000

    inputfile.close()
 
with open("15_fieldmap.json", "w") as output:
    json.dump(fieldmap, output)

#sys.exit(0)

input_directory = "IMSLP_neu"
output_filename = "15_output.mrc"

if len(sys.argv) > 1:
    input_directory = sys.argv[1]
if len(sys.argv) > 2:
    output_filename = sys.argv[2]

outputfile = io.open(output_filename, "wb")

failed = 0

for root, _, files in os.walk(input_directory):
    for filename in files:
        if not filename.endswith(".xml"):
            continue
        filepath = os.path.join(root, filename)
        inputfile = io.open(filepath, "r", encoding="utf-8")        

        f650a = []

        record = xmltodict.parse(inputfile.read())               
        record = record["document"]

        try:
            title = record["title"]
            #title = record["subject"]
        except:
            continue

        marcrecord = marcx.Record(force_utf8=True)
        marcrecord.strict = False
        
        # Leader
        marcrecord.leader = "     ncs  22        450 "

        # Identifikator
        f001 = record["identifier"]["#text"]
        marcrecord.add("001", data="finc-15-%s" % f001)
        
        # Format
        marcrecord.add("007", data="cr")
               
        # Sprache
        # besser: aus vifaxml parsen!!
        f041a = get_field("languages")
        marcrecord.add("008", data="130227uu20uuuuuuxx uuup%s  c" % f041a)
        marcrecord.add("041", a=f041a)
     
     
        # Komponist
        f100a = record["creator"]["mainForm"]
        f1000 = fieldmap[f001].get("viaf", "")
        marcrecord.add("100", a=f100a, e="cmp", _0=f1000)

        # Werktitel
        f240a = fieldmap[f001].get("title", "")
        marcrecord.add("240", a=f240a)
    
        # Sachtitel
        f245a = record["title"]
        f245a = html_unescape(f245a)
        marcrecord.add("245", a=f245a)
      
        # Alternativtitel
        f246a = get_field("additionalTitle")
        f246a = html_unescape(f246a)
        marcrecord.add("246", a=f246a)
        
        # Erscheinungsjahr
        # nicht mehr enthalten 
          
        # Kompositionsjahr     
        year = get_field("date")
        marcrecord.add("260", c=year)
        marcrecord.add("650", y=year)
        
        # Fußnote
        f500a = get_field("abstract")
        marcrecord.add("500", a=f500a)
    
        # Stil / Epoche       
        try:
            subject = record["subject"]
        except:
            subject = ""
        if subject != "":
            if isinstance(subject, list):
                # [OrderedDict([('mainForm', 'Piano piece')]), OrderedDict([('mainForm', 'Romantic')])]
                style = subject[0]["mainForm"]
                style = style.title()
                epoch = subject[1]["mainForm"] # Martin fragen, wegen "TypeError: string indices must be integers"
                epoch = epoch.title()
            else:
                epoch = record["subject"]["mainForm"]
                epoch = epoch.title()
            f650a.append(epoch)
            f590a = epoch
      
        # Besetzung
        instrumentation = get_field("music_arrangement_of")
        instrumentation = instrumentation.title()
        f650a.append(instrumentation)
        f590b = instrumentation

        marcrecord.add("590", subfields=["a", f590a, "b", f590b])

        # Schlagwörter       
        subtest = []
        for subject in f650a:
            subject = subject.title()
            if subject not in subtest:
                subtest.append(subject)
                marcrecord.add("689", a=subject)

        # Beitragende
        try:
            f700a = record["contributor"]["mainForm"]
            marcrecord.add("700", a=f700a, e="ctb")
        except:
            pass

        # URL
        f856u = record["url"]["#text"]
        marcrecord.add("856", q="text/html", _3="Petrucci Musikbibliothek", u=f856u)

        marcrecord.add("935", b="cofz", c="muno")

        marcrecord.add("970", c="PN")

        marcrecord.add("980", a=f001, b="15", c="Petrucci Musikbibliothek")

        outputfile.write(marcrecord.as_marc())

        inputfile.close()


outputfile.close()


# coding: utf-8
# pylint: disable=C301,C0103,W0201

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Provides a basic benchmark decorator. Usage:

    @timed
    def run():
        print('Running...')

Just logs the output. Could be extended to send this information to some
service if available.
"""

import functools
import logging
from builtins import object
from timeit import default_timer

logger = logging.getLogger('gluish')


class bcolors(object):
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def red(s):
    return bcolors.FAIL + s + bcolors.ENDC


def yellow(s):
    return bcolors.WARNING + s + bcolors.ENDC


def green(s):
    return bcolors.OKGREEN + s + bcolors.ENDC


class Timer(object):
    """
    A timer as a context manager, slight visual language.
    """

    def __init__(self, green=10, yellow=60):
        """
        Indicate runtimes with colors, green < 10s, yellow < 60s.
        """
        # measures wall clock time, not CPU time!
        # On Unix systems, it corresponds to time.time
        # On Windows systems, it corresponds to time.clock
        self.timer = default_timer

        self.green = green
        self.yellow = yellow

    def __enter__(self):
        self.start = self.timer()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.end = self.timer()
        self.elapsed_s = self.end - self.start
        self.elapsed_ms = self.elapsed_s * 1000


def timed(method):
    """
    A @timed decorator.
    """
    @functools.wraps(method)
    def _timed(*args, **kwargs):
        """
        Benchmark decorator.
        """
        with Timer() as timer:
            result = method(*args, **kwargs)
        klass = args[0].__class__.__name__
        fun = method.__name__

        msg = '[%s.%s] %0.5f' % (klass, fun, timer.elapsed_s)
        if timer.elapsed_s <= timer.green:
            logger.debug(green(msg))
        elif timer.elapsed_s <= timer.yellow:
            logger.debug(yellow(msg))
        else:
            logger.debug(red(msg))
        return result
    return _timed



# coding: utf-8
# pylint: disable=C0301,C0330,W0622,E1101

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""

CrossRef is an association of scholarly publishers
that develops shared infrastructure to support
more effective scholarly communications.

Our citation-linking network today covers over 68
million journal articles and other content items
(books chapters, data, theses, technical reports)
from thousands of scholarly and professional
publishers around the globe.

Configuration
-------------

[crossref]

doi-blacklist = /tmp/siskin-data/crossref/CrossrefDOIBlacklist/output.tsv

"""

import datetime
import io
import itertools
import os
import socket
import tempfile
import time
import urllib.error
import urllib.parse
import urllib.request
from builtins import range

import elasticsearch
import luigi
import requests
from future import standard_library

import ujson as json
from gluish.common import Executable
from gluish.format import TSV, Gzip
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import date_range, shellout
from siskin import __version__
from siskin.benchmark import timed
from siskin.mail import send_mail
from siskin.sources.amsl import AMSLFilterConfig, AMSLService
from siskin.task import DefaultTask
from siskin.utils import URLCache

standard_library.install_aliases()


class CrossrefTask(DefaultTask):
    """
    Crossref related tasks. See: http://www.crossref.org/
    """
    TAG = '49'

    def closest(self):
        """
        Update frequency.
        """
        return monthly(date=self.date)


class CrossrefHarvestChunkWithCursor(CrossrefTask):
    """
    Harvest window with cursors (https://git.io/v1K27).
    """
    begin = luigi.DateParameter(description='start of harvesting window')
    end = luigi.DateParameter(
        description='end of harvesting window, inclusive')
    filter = luigi.Parameter(
        default='deposit', description='index, deposit, update')

    rows = luigi.IntParameter(default=1000, significant=False)
    max_retries = luigi.IntParameter(
        default=10, significant=False, description='HTTP retries')
    attempts = luigi.IntParameter(
        default=3, significant=False, description='number of attempts to GET an URL that failed')
    sleep = luigi.IntParameter(
        default=1, significant=False, description='sleep between requests')

    def run(self):
        """
        > Using large offset values can result in extremely long response times. Offsets
        in the 100,000s and beyond will likely cause a timeout before the API is able
        to respond. An alternative to paging through very large result sets (like a
        corpus used for text and data mining) it to use the API's exposure of Solr's
        deep paging cursors. Any combination of query, filters and facets may be used
        with deep paging cursors. While rows may be specified along with cursor,
        offset and sample cannot be used. To use deep paging make a query as normal,
        but include the cursor parameter with a value of * (https://git.io/vFyAn).

        > But we prefer carrots to sticks. As of September 18th 2017 any API
        queries that use HTTPS and have appropriate contact information will be
        directed to a special pool of API machines that are reserved for polite
        users. (https://git.io/vFyN5), refs #9059.
        """
        cache = URLCache(directory=os.path.join(
            tempfile.gettempdir(), '.urlcache'))
        adapter = requests.adapters.HTTPAdapter(max_retries=self.max_retries)
        cache.sess.mount('http://', adapter)

        filter = "from-{self.filter}-date:{self.begin},until-{self.filter}-date:{self.end}".format(
            self=self)
        rows, offset = self.rows, 0

        cursor = '*'

        with self.output().open('w') as output:
            params = {'rows': rows, 'filter': filter}

            while True:
                params['cursor'] = cursor

                # Do not fail, if user has not configured mailto, https://git.io/vFyN5.
                if self.config.get('crossref', 'mailto', fallback=None):
                    params['mailto'] = self.config.get('crossref', 'mailto')

                url = 'https://api.crossref.org/works?%s' % (
                    urllib.parse.urlencode(params))

                for attempt in range(1, self.attempts):
                    if not cache.is_cached(url):
                        time.sleep(self.sleep)
                    body = cache.get(url)
                    try:
                        content = json.loads(body)
                    except ValueError as err:
                        if attempt == self.attempts - 1:
                            self.logger.debug('URL was %s', url)
                            self.logger.debug(err)
                            self.logger.debug(body[:100] + '...')
                            raise

                        cache_file = cache.get_cache_file(url)
                        if os.path.exists(cache_file):
                            self.logger.debug(
                                'trying to recover by removing cached entry at %s', cache_file)
                            os.remove(cache_file)
                    else:
                        break  # all fine

                count = len(content['message']['items'])
                self.logger.debug("%s: %s", url, count)
                if count == 0:
                    break

                output.write(body + '\n')
                offset += rows

                if not 'next-cursor' in content['message']:
                    raise RuntimeError(
                        'expected next-cursor in message, but missing')
                cursor = content['message']['next-cursor']

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))


class CrossrefHarvest(luigi.WrapperTask, CrossrefTask):
    """
    Harvest everything in incremental steps. Yield the targets sorted by date (ascending).
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    end = luigi.DateParameter(default=datetime.date.today())
    update = luigi.Parameter(
        default='months', description='days, weeks or months')

    def requires(self):
        if self.update not in ('days', 'weeks', 'months'):
            raise RuntimeError('update can only be: days, weeks or months')
        dates = [dt for dt in date_range(self.begin, self.end, 1, self.update)]
        tasks = [CrossrefHarvestChunkWithCursor(begin=dates[i], end=dates[i + 1])
                 for i in range(len(dates) - 1)]
        return sorted(tasks)

    def output(self):
        return self.input()


class CrossrefChunkItems(CrossrefTask):
    """
    Extract the message items, per chunk.
    """
    begin = luigi.DateParameter()
    end = luigi.DateParameter()
    filter = luigi.Parameter(
        default='deposit', description='index, deposit, update')

    def requires(self):
        return CrossrefHarvestChunkWithCursor(begin=self.begin, end=self.end, filter=self.filter)

    def run(self):
        output = shellout(
            "jq -c -r '.message.items[]?' {input} | pigz -c > {output}", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class CrossrefRawItems(CrossrefTask):
    """
    Concatenate all harvested items.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())
    update = luigi.Parameter(
        default='months', description='days, weeks or months')

    def requires(self):
        if self.update not in ('days', 'weeks', 'months'):
            raise RuntimeError('update can only be: days, weeks or months')
        dates = [dt for dt in date_range(
            self.begin, self.date, 1, self.update)]
        tasks = [CrossrefChunkItems(begin=dates[i - 1], end=dates[i])
                 for i in range(1, len(dates))]
        return tasks

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for target in self.input():
            shellout("cat {input} >> {output}",
                     input=target.path, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class CrossrefUniqItems(CrossrefTask):
    """
    Calculate current snapshot via span-crossref-snapshot.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return CrossrefRawItems(begin=self.begin, date=self.closest())

    def run(self):
        output = shellout(
            "span-crossref-snapshot -verbose -z -o {output} {input}", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class CrossrefIntermediateSchema(CrossrefTask):
    """
    Convert to intermediate format via span.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'span': Executable(name='span-import', message='http://git.io/vI8NV'),
            'file': CrossrefUniqItems(begin=self.begin, date=self.date)
        }

    @timed
    def run(self):
        output = shellout("span-import -i crossref <(unpigz -c {input}) | pigz -c > {output}",
                          input=self.input().get('file').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))


class CrossrefCollectionStats(CrossrefTask):
    """
    Export a list of collection name and a few fields.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return CrossrefIntermediateSchema(begin=self.begin, date=self.date)

    @timed
    def run(self):
        output = shellout("""
            unpigz -c {input} |
            jq -rc '[.["finc.id"], .["finc.source_id"], .["finc.mega_collection"], .["doi"]] | @csv' | sed -e 's@","@\t@g' | tr -d '"' > {output}
        """, input=self.input().path, preserve_whitespace=True)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())


class CrossrefExport(CrossrefTask):
    """
    Tag with ISILs, then export to various formats.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='solr5vu3', description='export format')

    def requires(self):
        return {
            'file': CrossrefIntermediateSchema(date=self.date),
            'config': AMSLFilterConfig(date=self.date),
        }

    def run(self):
        output = shellout("span-tag -c {config} <(unpigz -c {input}) | pigz -c > {output}",
                          config=self.input().get('config').path, input=self.input().get('file').path)
        output = shellout(
            "span-export -o {format} <(unpigz -c {input}) | pigz -c > {output}", format=self.format, input=output)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        extensions = {
            'solr5vu3': 'ldj.gz',
            'formeta': 'form.gz',
        }
        return luigi.LocalTarget(path=self.path(ext=extensions.get(self.format, 'gz')))


class CrossrefCollections(CrossrefTask):
    """
    A collection of crossref collections, refs. #6985. XXX: Save counts as well.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'input': CrossrefIntermediateSchema(begin=self.begin, date=self.date),
                'jq': Executable(name='jq', message='https://github.com/stedolan/jq')}

    @timed
    def run(self):
        output = shellout("""jq -rc '.["finc.mega_collection"][]?' <(unpigz -c {input}) | LC_ALL=C sort -S35% -u > {output}""",
                          input=self.input().get('input').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefCollectionsCount(CrossrefTask):
    """
    Report collections and the number of titles per collection.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'input': CrossrefIntermediateSchema(begin=self.begin, date=self.date),
                'jq': Executable(name='jq', message='https://github.com/stedolan/jq')}

    @timed
    def run(self):
        output = shellout("""jq -rc '.["finc.mega_collection"][]?' <(unpigz -c {input}) | LC_ALL=C sort -S35% > {output}""",
                          input=self.input().get('input').path)

        groups = {}  # Map collection name to its size.
        with open(output) as handle:
            for k, g in itertools.groupby(handle):
                name = k.strip()
                groups[name] = len(list(g))

        with self.output().open('w') as output:
            json.dump(groups, output)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='json'))


class CrossrefCollectionsDifference(CrossrefTask):
    """
    Refs. #7049. Check list of collections against AMSL Crossref collections and
    report difference.

    This task uses an experimental email template in ../assets/mail/7049.tmpl, by
    default no email is sent, only when --to is set to one or more comma
    separated email addresses.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    to = luigi.Parameter(default=None, description="email address of recipient, comma separated, if multiple")

    def requires(self):
        return {
            'crossref': CrossrefCollections(begin=self.begin, date=self.date),
            'amsl': AMSLService(date=self.date, name='outboundservices:discovery')
        }

    @timed
    def run(self):
        if self.to is None:
            self.logger.debug("not sending any email, use --to my@mail.com to send out a report")

        amsl = set()

        with self.input().get('amsl').open() as handle:
            items = json.load(handle)

        for item in items:
            if item['sourceID'] == '49':
                amsl.add(item['megaCollection'].strip())

        self.logger.debug("found %s crossref collections in AMSL" % len(amsl))

        missing_in_amsl = []

        with self.input().get('crossref').open() as handle:
            with self.output().open('w') as output:
                for row in handle.iter_tsv(cols=('name',)):
                    if row.name not in amsl:
                        missing_in_amsl.append(row.name)
                        output.write_tsv(row.name)

        self.logger.debug("%s collections seem to be missing in AMSL", len(missing_in_amsl))

        if self.to is not None:
            # Try to send a message, experimental.
            tolist = [v.strip() for v in self.to.split(",")]

            subject = "%s %s %s" % (self.__class__.__name__,
                                    datetime.datetime.today().strftime("%Y-%m-%d %H:%M"),
                                    len(missing_in_amsl))

            with io.open(self.assets("mail/7049.tmpl"), encoding="utf-8") as fh:
                template = fh.read()

            message = template.format(
                version=__version__,
                count=len(missing_in_amsl),
                clist="\n".join(missing_in_amsl),
                hostname=socket.gethostname(),
                xref=self.input().get("crossref").path,
                amsl=self.input().get("amsl").path,
            )
            message = message.encode("utf-8")
            send_mail(tolist=tolist, subject=subject, message=message)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefDOIList(CrossrefTask):
    """
    A list of Crossref DOIs.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'input': CrossrefIntermediateSchema(date=self.date),
                'jq': Executable(name='jq', message='https://github.com/stedolan/jq')}

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        # process substitution sometimes results in a broken pipe, so extract beforehand
        output = shellout(
            "unpigz -c {input} > {output}", input=self.input().get('input').path)
        shellout("""jq -r '.doi?' {input} | grep -o "10.*" 2> /dev/null | LC_ALL=C sort -S50% > {output} """,
                 input=output, output=stopover)
        os.remove(output)
        output = shellout(
            """sort -S50% -u {input} > {output} """, input=stopover)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefISSNList(CrossrefTask):
    """
    Just dump a list of all ISSN values. With dups and all.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'input': CrossrefUniqItems(begin=self.begin, date=self.date),
                'jq': Executable(name='jq', message='https://github.com/stedolan/jq')}

    @timed
    def run(self):
        output = shellout(
            "jq -r '.ISSN[]?' <(unpigz -c {input}) 2> /dev/null > {output}", input=self.input().get('input').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefUniqISSNList(CrossrefTask):
    """
    Just dump a list of all ISSN values. Sorted and uniq.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return CrossrefISSNList(begin=self.begin, date=self.date)

    @timed
    def run(self):
        output = shellout(
            "sort -u {input} > {output}", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefDOIAndISSNList(CrossrefTask):
    """
    A list of Crossref DOIs with their ISSNs.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'input': CrossrefIntermediateSchema(date=self.date),
                'jq': Executable(name='jq', message='https://github.com/stedolan/jq')}

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        temp = shellout("unpigz -c {input} > {output}",
                        input=self.input().get('input').path)
        output = shellout("""jq -r '[.doi?, .["rft.issn"][]?, .["rft.eissn"][]?] | @csv' {input} | LC_ALL=C sort -S50% > {output} """,
                          input=temp, output=stopover)
        os.remove(temp)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='csv'))


class CrossrefHarvestGeneric(CrossrefTask):
    """
    Basic harvest of members, funders, etc.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='members')

    rows = luigi.IntParameter(default=1000, significant=False)
    max_retries = luigi.IntParameter(default=10, significant=False)

    @timed
    def run(self):
        cache = URLCache(directory=os.path.join(
            tempfile.gettempdir(), '.urlcache'))
        adapter = requests.adapters.HTTPAdapter(max_retries=self.max_retries)
        cache.sess.mount('http://', adapter)

        rows, offset = self.rows, 0

        with self.output().open('w') as output:
            while True:
                params = {"rows": rows, "offset": offset}
                url = 'http://api.crossref.org/%s?%s' % (
                    self.kind, urllib.parse.urlencode(params))
                for attempt in range(1, 3):
                    body = cache.get(url)
                    try:
                        content = json.loads(body)
                    except ValueError as err:
                        if attempt == 2:
                            self.logger.debug("URL was %s", url)
                            self.logger.debug(err)
                            self.logger.debug(body[:100])
                            raise
                        if os.path.exists(cache.get_cache_file(url)):
                            self.logger.debug(
                                "trying to recover by removing cached entry")
                            os.remove(cache.get_cache_file(url))
                    else:
                        break
                items = content["message"]["items"]
                self.logger.debug("%s: %s", url, len(items))
                if len(items) == 0:
                    break
                output.write(body)
                output.write("\n")
                offset += rows

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))


class CrossrefGenericItems(CrossrefTask):
    """
    Flatten and deduplicate. Stub.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='members')

    def requires(self):
        return CrossrefHarvestGeneric(begin=self.begin, date=self.closest(), kind=self.kind)

    @timed
    def run(self):
        with self.output().open('w') as output:
            with self.input().open() as handle:
                for line in handle:
                    content = json.loads(line)
                    if not content.get("status") == "ok":
                        raise RuntimeError("invalid response status")
                    items = content["message"]["items"]
                    for item in items:
                        output.write(json.dumps(item))
                        output.write("\n")

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))


class CrossrefDOIHarvest(CrossrefTask):
    """
    Harvest DOI redirects from doi.org API. This is a long running task. It's
    probably best to run this tasks complete separate from the rest and let
    other processing pipelines use the result, as it is available.

    ----

    It is highly recommended that you put a static entry into /etc/hosts
    for doi.org while `hurrly` is running.

    As of 2015-07-31 doi.org resolves to six servers. Just choose one.

    $ nslookup doi.org

    """

    def requires(self):
        """
        If we have more DOI sources, we could add them as requirements here.
        """
        return {
            'input': CrossrefDOIList(date=datetime.date.today()),
            'hurrly': Executable(name='hurrly', message='http://github.com/miku/hurrly'),
            'pigz': Executable(name='pigz', message='http://zlib.net/pigz/')
        }

    def run(self):
        output = shellout(
            "hurrly -w 4 < {input} | pigz > {output}", input=self.input().get('input').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='tsv.gz'))


class CrossrefDOIBlacklist(CrossrefTask):
    """
    Create a blacklist of DOIs. Possible cases:

    1. A DOI redirects to http://www.crossref.org/deleted_DOI.html or
       most of these sites below crossref.org: https://gist.github.com/miku/6d754104c51fb553256d
    2. A DOI API lookup does not return a HTTP 200.

    The output of this task should be used as doi-blacklist in config.
    """

    def requires(self):
        return CrossrefDOIHarvest()

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        shellout("""LC_ALL=C zgrep -E "http(s)?://.*.crossref.org" {input} >> {output}""",
                 input=self.input().path, output=stopover)
        shellout("""LC_ALL=C zgrep -v "^200" {input} >> {output}""",
                 input=self.input().path, output=stopover)
        output = shellout(
            "sort -S50% -u {input} | cut -f4 | sed s@http://doi.org/api/handles/@@g > {output}", input=stopover)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())



# coding: utf-8
# pylint: disable=C0301,E1101,C0103

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

#
# BTAG preparation
# ================

import collections
import json
import random

import luigi

from gluish.format import TSV
from siskin.database import sqlitedb
from siskin.sources.crossref import CrossrefUniqItems
from siskin.sources.doaj import DOAJDOIList, DOAJDump
from siskin.sources.mag import MAGReferenceDB
from siskin.task import DefaultTask


class BTAGTask(DefaultTask):
    """ Base task for BTAG. """
    TAG = 'btag'


class BTAGCrossrefSet(BTAGTask):
    """
    Find a set of crossref records with duplicates and references.
    """

    max = luigi.IntParameter(default=100000000)
    minrefs = luigi.IntParameter(default=2)

    def requires(self):
        return {
            'crossref': CrossrefUniqItems(),
            'doaj': DOAJDOIList(),
            'magdb': MAGReferenceDB()
        }

    def run(self):
        """
        Write out the doi, the number of refs found in MAG and whether the doi is in DOAJ as well.
        """
        doaj = set()
        with self.input().get('doaj').open() as handle:
            for line in handle:
                doaj.add(line.strip())

        with sqlitedb(self.input().get('magdb').path) as cursor:
            with self.input().get('crossref').open() as handle:
                with self.output().open('w') as output:
                    stmt = """
                    select doi from lookup where id IN (
                        select ref from refs where id = (
                        select id from lookup where doi = ?))
                    """
                    for i, line in enumerate(handle):
                        if i > self.max:
                            break
                        doc = json.loads(line)
                        doi = doc.get('DOI')
                        if doi is not None:
                            cursor.execute(stmt, (doi,))
                            rows = cursor.fetchall()
                            refs = [v[0] for v in rows]
                            if len(refs) >= self.minrefs:
                                output.write_tsv(doi, len(refs), doi in doaj)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class BTAGCrossref(BTAGTask):
    """
    Create a crossref subset, that contains items, that have references in MAG and dups in DOAJ.
    """

    def requires(self):
        return {
            'crossref': CrossrefUniqItems(),
            'set': BTAGCrossrefSet(max=5000000),
        }

    def run(self):
        entries = {}
        with self.input().get('set').open() as handle:
            for row in handle.iter_tsv(cols=('doi', 'refcount', 'indoaj')):
                entries[row.doi] = row

        with self.input().get('crossref').open() as handle:
            with self.output().open('w') as output:
                for line in handle:
                    doc = json.loads(line)
                    if doc.get("DOI") in entries:
                        output.write(json.dumps(doc) + "\n")
                    else:
                        if random.random() < 0.0001:
                            self.logger.debug("inlude random record")
                            output.write(json.dumps(doc) + "\n")

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))


class BTAGDOAJSubset(BTAGTask):
    """
    A smaller subset of DOAJ.
    """

    def requires(self):
        return {
            'doaj': DOAJDump(),
            'set': BTAGCrossrefSet(max=5000000),
        }

    def run(self):
        entries = {}
        with self.input().get('set').open() as handle:
            for row in handle.iter_tsv(cols=('doi', 'refcount', 'indoaj')):
                entries[row.doi] = row

        with self.input().get('doaj').open() as handle:
            with self.output().open('w') as output:
                for line in handle:
                    doc = json.loads(line)
                    for iddoc in doc['_source']['bibjson']['identifier']:
                        if iddoc["type"] != "doi":
                            continue
                        doi = iddoc["id"]
                        if doi in entries:
                            output.write(json.dumps(doc) + "\n")
                        else:
                            if random.random() < 0.001:
                                self.logger.debug("inlude random record")
                                output.write(json.dumps(doc) + "\n")

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))


class BTAGMAGSubset(BTAGTask):
    """ A small subset of MAG. """

    def requires(self):
        return {
            'db': MAGReferenceDB(),
            'crossref': BTAGCrossref(),
        }

    def run(self):
        doiset = set()
        with self.input().get('crossref').open() as handle:
            for line in handle:
                doc = json.loads(line)
                doiset.add(doc['DOI'])

        with sqlitedb(self.input().get('db').path) as cursor:
            with self.output().open('w') as output:
                stmt = """
                    select doi from lookup where id IN (
                    select ref from refs where id = (
                        select id from lookup where doi = ?))
                """
                for doi in doiset:
                    cursor.execute(stmt, (doi,))
                    rows = cursor.fetchall()
                    refs = [v[0] for v in rows]
                    output.write_tsv(doi, *refs)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'), format=TSV)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Robert Schenk, <robert.schenk@uni-leipzig.de>
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
VKFilm Bundesarchiv, #8695.

TODO: Unify VKFilm* tasks.

Config
------

[vkfilmba]

baseurl = https://example.com

data = http://example.com/data.zip
deletions = http://example.com/del.txt

"""

import datetime
import hashlib
import os
import tempfile

import luigi

from gluish.format import TSV
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.common import FTPMirror
from siskin.task import DefaultTask


class VKFilmBATask(DefaultTask):
    TAG = '148'

    # As this is a manual process, adjust this list by hand and rerun tasks.
    filenames = [
        "adlr.zip",
        "adlr_171011.zip",
        "adlr_1712.zip",
    ]

    def fingerprint(self):
        """
        Generate a name based on the inputs, so everytime we adjust the file
        list, we get another output.
        """
        h = hashlib.sha1()
        h.update("".join(self.filenames).encode("utf-8"))
        return h.hexdigest()


class VKFilmBADownload(VKFilmBATask):
    """
    Download raw data daily. At least try, since as of 2018-02-15, file seems
    missing, refs #12460.
    """
    date = luigi.DateParameter(default=datetime.date.today())

    def run(self):
        """
        XXX: Maybe check, if we already downloaded this.
        """
        output = shellout("""curl --fail "{url}" > {output} """,
                          url=self.config.get('vkfilmba', 'data'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        """
        We expect zip, but to guarantee.
        """
        return luigi.LocalTarget(path=self.path(ext="zip"))


class VKFilmBADownloadDeletions(VKFilmBATask):
    """
    Download raw deletions daily. At least try, since as of 2018-02-15, file seems
    missing, refs #12460.
    """
    date = luigi.DateParameter(default=datetime.date.today())

    def run(self):
        output = shellout("""curl --fail "{url}" > {output} """,
                          url=self.config.get('vkfilmba', 'deletions'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        """
        We expect txt, but to guarantee.
        """
        return luigi.LocalTarget(path=self.path(ext="txt"))


class VKFilmBADump(VKFilmBATask):
    """
    Concatenate a list of URLs.

    Note: DO NOT DELETE the output of this task. There are copies:

    * https://goo.gl/FvhXnL
    * https://speicherwolke.uni-leipzig.de/index.php/s/KMUldvGMJRc7iLP

    XXX: There is a "delete-list" of ID, which should be filtered here.
    """

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for fn in self.filenames:
            link = os.path.join(self.config.get("vkfilmba", "baseurl"), fn)
            output = shellout("""wget -O {output} "{link}" """, link=link)
            output = shellout(""" unzip -p "{input}" > "{output}" """,
                              input=output)
            output = shellout("""yaz-marcdump -i marcxml -o marc "{input}" > "{output}"  """,
                              input=output, ignoremap={5: "Fixme."})
            shellout("""cat "{input}" >> "{stopover}" """,
                     input=output, stopover=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(filename="%s.mrc" % self.fingerprint()))


class VKFilmBAConvert(VKFilmBATask):
    """
    Convert download, refs #12460.
    """

    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'dump': VKFilmBADump(),
            'download': VKFilmBADownload(),
        }

    def run(self):
        # XXX: workaround; we want a single file for further processing.
        filenames = [
            'adlr_1m1',
            'adlr_1801',
            'adlr',
        ]
        _, stopover = tempfile.mkstemp(prefix='siskin-')

        for fn in filenames:
            output = shellout(""" unzip -p "{input}" {fn} > "{output}" """, input=self.input().get('download').path, fn=fn)
            output = shellout("""yaz-marcdump -i marcxml -o marc "{input}" > "{output}"  """,
                              input=output, ignoremap={5: "Fixme."})
            shellout("""cat "{input}" >> "{stopover}" """,
                     input=output, stopover=stopover)

        # Concatenate, since download was only 150k, while dump was 9M?
        output = shellout("cat {a} {b} > {output}", a=self.input().get('dump').path, b=stopover)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(filename="%s.mrc" % self.fingerprint()))


class VKFilmBAMARC(VKFilmBATask):
    """
    Run conversion script.
    """

    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return VKFilmBAConvert()  # return VKFilmBADump()

    def run(self):
        output = shellout("python {script} {input} {output}",
                          script=self.assets("148/148_marcbinary.py"),
                          input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='fincmarc.mrc'))



# coding: utf-8
# pylint: disable=C0103,C0301

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Define version, disable some warnings and ensure temporary and log directories
are there and writeable.
"""

from __future__ import print_function

import os
import sys
import tempfile
import warnings

from siskin.configuration import Config

warnings.filterwarnings("ignore")

# https://urllib3.readthedocs.org/en/latest/security.html#insecurerequestwarning
# wrap into try-except, since at install-time urllib3 might not be installed yet
try:
    import urllib3
    urllib3.disable_warnings()
except (AttributeError, ImportError):
    pass

__version__ = '0.24.0'

config = Config.instance()
if sys.version_info.major == 2:
    tempfile.tempdir = config.get('core', 'tempdir')
else:
    tempfile.tempdir = config.get(
        'core', 'tempdir', fallback=tempfile.gettempdir())



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Springer (testing).

Options currently: Download via OwnCloud and (preferred) SFTP.

Config
------

[springer]

username = admin
password = s3cret
url = http://export.com/intermediate.file.gz

ftp-host = sftp://example.com
ftp-base = /target_data/xyz
ftp-username= admin
ftp-password= s3cr3t
ftp-pattern = *

"""

from __future__ import print_function

import datetime
import re

import luigi
import six
import ujson as json
from gluish.format import TSV, Gzip
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout

from siskin.common import FTPMirror
from siskin.decorator import deprecated
from siskin.sources.amsl import AMSLFilterConfig
from siskin.task import DefaultTask


class SpringerTask(DefaultTask):
    TAG = '105'

    def closest(self):
        return weekly(date=self.date)


class SpringerPaths(SpringerTask):
    """
    Mirror SLUB FTP. Preferred as of November 2017.
    """

    date = ClosestDateParameter(default=datetime.date.today())
    max_retries = luigi.IntParameter(default=10, significant=False)
    timeout = luigi.IntParameter(
        default=20, significant=False, description='timeout in seconds')

    def requires(self):
        return FTPMirror(host=self.config.get('springer', 'ftp-host'),
                         base=self.config.get('springer', 'ftp-base'),
                         username=self.config.get('springer', 'ftp-username'),
                         password=self.config.get('springer', 'ftp-password'),
                         pattern=self.config.get('springer', 'ftp-pattern'),
                         max_retries=self.max_retries,
                         timeout=self.timeout)

    def run(self):
        self.input().move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class SpringerDownload(SpringerTask):
    """
    Attempt to download file from a configured URL. Preferred up until October 2017.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    @deprecated
    def run(self):
        output = shellout(""" curl --fail -v -u {username}:{password} "{url}" > {output} """,
                          username=self.config.get('springer', 'username'),
                          password=self.config.get('springer', 'password'),
                          url=self.config.get('springer', 'url'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class SpringerIssue11557(SpringerTask):
    """
    Via 5994#note-37, finc.mega_collection is now multi-valued, refs #11557.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return SpringerDownload(date=self.date)

    @deprecated
    def run(self):
        output = shellout("""
            jq -rc 'del(.["finc.AIRecordType"]) | del(.["AIAccessFacet"]) | .["finc.mega_collection"] = [.["finc.mega_collection"]]' < <(unpigz -c {input}) | pigz -c > {output}
        """, input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class SpringerCleanup(SpringerTask):
    """
    2017-11-28: finc.mega_collection is now multi-valued; AIAccessFacet remains.
    2017-12-12: new finc.id, refs #11821, #11960, #11961.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return SpringerPaths(date=self.date)

    def run(self):
        realpath = None
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                if not row.path.endswith("total_tpu.ldj.gz"):
                    continue
                realpath = row.path
                break
            else:
                raise RuntimeError('FTP site does not contain total_tpu.ldj.gz')
        output = shellout("""
            unpigz -c {input} | jq -rc 'del(.["finc.AIRecordType"]) | del(.["AIAccessFacet"])' |
                jq -c '. + {{"finc.id": .["finc.record_id"], "finc.record_id": .doi}}' | pigz -c > {output}
        """, input=realpath)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class SpringerCleanFields(SpringerTask):
    """
    As per Google Hangout 2017-09-08 we define "exchange-ready" as
    "no-post-processing-required". Hence this task marked deprecated, use
    SpringerDownload task instead.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return SpringerDownload(date=self.date)

    @deprecated
    def run(self):
        striptags = lambda s: re.sub(r'\$\$[^\$]*\$\$', '', re.sub(r'<[^>]*>', '', s))
        with self.input().open() as handle:
            with self.output().open('w') as output:
                for line in handle:
                    doc = json.loads(line)
                    doc['abstract'] = striptags(doc.get('abstract', '').encode('utf-8'))
                    doc['rft.atitle'] = striptags(doc.get('rft.atitle', '').encode('utf-8'))
                    doc['x.subjects'] = [striptags(subj.encode('utf-8')) for subj in doc.get('x.subjects', [])]

                    # #5994, #note-37 / https://git.io/v5ZNu requests
                    # multi-valued mega_collection, rev: https://git.io/v5ZNl.
                    # Turn value into list, if it is not already.
                    if isinstance(doc['finc.mega_collection'], six.string_types):
                        doc['finc.mega_collection'] = [doc['finc.mega_collection']]

                    output.write(json.dumps(doc))
                    output.write("\n")

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class SpringerIntermediateSchema(SpringerTask, luigi.WrapperTask):
    """
    Just the cleaned version.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return SpringerCleanup(date=self.date)

    def output(self):
        return self.input()



# coding: utf-8
# pylint: disable=C0301

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Common tasks.
"""

import datetime
import email.utils as eut
import hashlib
import json
import os
import pipes
import tempfile

import luigi
import requests
from gluish.common import Executable
from gluish.format import TSV
from gluish.utils import shellout

from siskin.task import DefaultTask
from siskin.utils import iterfiles, random_string


class CommonTask(DefaultTask):
    """
    A base class for common classes. These artefacts will be written to the
    systems tempdir.
    """
    TAG = 'common'


class FTPMirror(CommonTask):
    """
    A generic FTP directory sync. Requires lftp (http://lftp.yar.ru/).  The
    output of this task is a single file, that contains the paths to all the
    mirrored files.
    """
    host = luigi.Parameter()
    username = luigi.Parameter(default='anonymous')
    password = luigi.Parameter(default='')
    pattern = luigi.Parameter(default='*', description="e.g. '*leip_*.zip'")
    base = luigi.Parameter(default='.')
    indicator = luigi.Parameter(default=random_string())
    max_retries = luigi.IntParameter(default=5, significant=False)
    timeout = luigi.IntParameter(
        default=10, significant=False, description='timeout in seconds')
    exclude_glob = luigi.Parameter(
        default="", significant=False, description='globs to exclude')

    def requires(self):
        return Executable(name='lftp', message='http://lftp.yar.ru/')

    def run(self):
        """ The indicator is always recreated, while the subdir
        for a given (host, username, base, pattern) is just synced. """
        base = os.path.dirname(self.output().path)
        subdir = hashlib.sha1('{host}:{username}:{base}:{pattern}'.format(
            host=self.host, username=self.username, base=self.base,
            pattern=self.pattern).encode('utf-8')).hexdigest()

        # target is the root of the mirror
        target = os.path.join(base, subdir)
        if not os.path.exists(target):
            os.makedirs(target)

        exclude_glob = ""
        if not self.exclude_glob == "":
            exclude_glob = "--exclude-glob %s" % self.exclude_glob

        command = """lftp -u {username},{password}
        -e "
            set sftp:auto-confirm yes;
            set net:max-retries {max_retries};
            set net:timeout {timeout};
            set mirror:parallel-directories 1;
            set ssl:verify-certificate no;
            set ftp:ssl-protect-data true;

        mirror --verbose=0 --only-newer {exclude_glob} -I {pattern} {base} {target}; exit" {host}"""

        shellout(command, host=self.host, username=pipes.quote(self.username),
                 password=pipes.quote(self.password),
                 pattern=pipes.quote(self.pattern),
                 target=pipes.quote(target),
                 base=pipes.quote(self.base),
                 max_retries=self.max_retries,
                 timeout=self.timeout,
                 exclude_glob=exclude_glob)

        with self.output().open('w') as output:
            for path in iterfiles(target):
                self.logger.debug("Mirrored: %s", path)
                output.write_tsv(path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True), format=TSV)


class HTTPDownload(CommonTask):
    """
    Download a file via HTTP, read out the HTTP Last-modified header and use it as filename.
    """
    url = luigi.Parameter(description='pass this parameter')

    def filename(self):
        """
        Returns the name of the output file. This helper relies on network connectivity.
        """
        r = requests.head(self.url)
        if r.status_code != 200:
            raise RuntimeError('%s on %s' % (r.status_code, self.url))
        value = r.headers.get('Last-Modified')
        if value is None:
            raise RuntimeError(
                'HTTPDownload relies on HTTP Last-Modified header at the moment')
        parsed_date = eut.parsedate(value)
        if parsed_date is None:
            raise RuntimeError('could not parse Last-Modifier header')
        last_modified_date = datetime.date(*parsed_date[:3])
        digest = hashlib.sha1(self.url).hexdigest()
        return '%s-%s.file' % (digest, last_modified_date.isoformat())

    def run(self):
        """
        We try just once. TODO(miku): Some retry bracket.
        Last-Modified date format: Wed, 25 Jan 2017 14:04:59 GMT
        """
        output = shellout(
            """ curl --fail "{url}" > {output} """, input=self.url)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.filename())


class RedmineDownload(CommonTask):
    """
    Download issue from Redmine via API.

    Requires config entry:

    [redmine]

    baseurl = https://projects.examples.com
    apikey = 123123123-456ABC
    """
    issue = luigi.Parameter(description="issue number")
    date = luigi.DateParameter(default=datetime.date.today())

    def run(self):
        self.logger.info("Accessing Redmine Issue #%s (%s/issues/%s) ...",
                         self.issue, self.config.get('redmine', 'baseurl'), self.issue)
        url = "%s/issues/%s.json?include=attachments" % (
            self.config.get('redmine', 'baseurl'), self.issue)
        output = shellout(""" curl -vL --fail -H "X-Redmine-API-Key:{apikey}" "{url}" > {output}""",
                          apikey=self.config.get("redmine", "apikey"), url=url, issue=self.issue)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="json"))


class RedmineDownloadAttachments(CommonTask):
    """
    Download all attachements for a ticket and make them temporarily accessible
    to other tasks. Redmine attachments are by default limited to about 10M, so
    it is ok to not cache anything here.

    Target is a file containing a list of filenames downloaded.
    """
    issue = luigi.Parameter(description="issue number")
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return RedmineDownload(issue=self.issue, date=self.date)

    def run(self):
        with self.input().open() as handle:
            doc = json.load(handle)
        tempdir = tempfile.mkdtemp(prefix='tmp-siskin-')
        for attachment in doc['issue']['attachments']:
            target = os.path.join(
                tempdir, os.path.basename(attachment["content_url"]))
            shellout("""curl -vL --fail -H "X-Redmine-API-Key:{apikey}" -o {target} "{url}" """,
                     url=attachment["content_url"], apikey=self.config.get(
                         "redmine", "apikey"),
                     target=target)

        with self.output().open('w') as output:
            for path in iterfiles(tempdir):
                self.logger.debug("Downloaded: %s", path)
                output.write_tsv(path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist'), format=TSV)
